import json
import zipfile
from pathlib import Path

from brokk_code.session_persistence import (
    get_session_zip_path,
    get_state_dir,
    has_tasks,
    load_last_session_id,
    save_last_session_id,
)


def test_state_dir_path():
    workspace = Path("/tmp/fake_ws")
    assert get_state_dir(workspace) == workspace / ".brokk"


def test_session_zip_path_creation(tmp_path):
    workspace = tmp_path
    session_id = "test-uuid-123"
    zip_path = get_session_zip_path(workspace, session_id)

    assert zip_path.name == "test-uuid-123.zip"
    assert zip_path.parent.name == "sessions"
    assert zip_path.parent.parent.name == ".brokk"
    # Ensure directory was created
    assert zip_path.parent.exists()


def test_last_session_id_roundtrip(tmp_path):
    workspace = tmp_path
    session_id = "8888-9999-aaaa"

    # Initially None
    assert load_last_session_id(workspace) is None

    # Save and Load
    save_last_session_id(workspace, session_id)
    assert load_last_session_id(workspace) == session_id


def test_load_last_session_id_tolerant(tmp_path):
    workspace = tmp_path
    state_dir = get_state_dir(workspace)
    state_dir.mkdir()
    session_file = state_dir / "last_session.json"

    # Malformed JSON
    session_file.write_text("invalid json {")
    assert load_last_session_id(workspace) is None

    # Missing key
    session_file.write_text(json.dumps({"wrongKey": "value"}))
    assert load_last_session_id(workspace) is None


def test_save_last_session_creates_missing_dirs(tmp_path):
    # Test path where .brokk doesn't exist yet
    workspace = tmp_path / "new_project"
    session_id = "new-uuid"

    save_last_session_id(workspace, session_id)
    assert (workspace / ".brokk" / "last_session.json").exists()
    assert load_last_session_id(workspace) == session_id


def test_has_tasks_missing_or_invalid(tmp_path):
    assert has_tasks(tmp_path / "nonexistent.zip") is False

    corrupt = tmp_path / "corrupt.zip"
    corrupt.write_text("not a zip")
    assert has_tasks(corrupt) is False


def test_has_tasks_history_definition(tmp_path):
    zip_path = tmp_path / "session.zip"

    def create_zip(lines: list[str] | None):
        with zipfile.ZipFile(zip_path, "w") as z:
            if lines is not None:
                z.writestr("contexts.jsonl", "\n".join(lines))

    # Missing contexts.jsonl
    with zipfile.ZipFile(zip_path, "w") as z:
        z.writestr("something_else.txt", "data")
    assert has_tasks(zip_path) is False

    # Empty contexts.jsonl
    create_zip([])
    assert has_tasks(zip_path) is False

    # contexts.jsonl but no tasks array
    create_zip([json.dumps({"id": "ctx1"})])
    assert has_tasks(zip_path) is False

    # tasks array but missing meta
    create_zip([json.dumps({"tasks": [{"sequence": 1}]})])
    assert has_tasks(zip_path) is False

    # tasks array with meta but missing/invalid sequence
    create_zip([json.dumps({"tasks": [{"taskType": "LUTZ"}]})])
    assert has_tasks(zip_path) is False

    # Success: meta and sequence
    create_zip([json.dumps({"tasks": [{"sequence": 1, "taskType": "LUTZ"}]})])
    assert has_tasks(zip_path) is True

    # Tolerant parsing: malformed line then valid line
    create_zip(
        ["{ invalid json", json.dumps({"tasks": [{"sequence": 1, "primaryModelName": "gpt-4o"}]})]
    )
    assert has_tasks(zip_path) is True

    # Check other meta fields
    create_zip([json.dumps({"tasks": [{"sequence": 2, "primaryModelReasoning": "logic"}]})])
    assert has_tasks(zip_path) is True

    # Success: single line without trailing newline
    with zipfile.ZipFile(zip_path, "w") as z:
        data = json.dumps({"tasks": [{"sequence": 1, "taskType": "LUTZ"}]}).encode("utf-8")
        z.writestr("contexts.jsonl", data)
    assert has_tasks(zip_path) is True


def test_has_tasks_respects_bounds(tmp_path, monkeypatch):
    from brokk_code import session_persistence

    # Set very low bounds for testing
    monkeypatch.setattr(session_persistence, "MAX_CONTEXTS_JSONL_BYTES", 100)
    monkeypatch.setattr(session_persistence, "MAX_CONTEXTS_JSONL_LINES", 3)

    zip_path = tmp_path / "bounded.zip"

    def create_zip(lines: list[str]):
        with zipfile.ZipFile(zip_path, "w") as z:
            z.writestr("contexts.jsonl", "\n".join(lines))

    # Case 1: Qualifying task only after the line cap
    create_zip(
        [
            json.dumps({"msg": "line1"}),
            json.dumps({"msg": "line2"}),
            json.dumps({"msg": "line3"}),
            json.dumps({"tasks": [{"sequence": 1, "taskType": "LUTZ"}]}),
        ]
    )
    assert has_tasks(zip_path) is False

    # Case 2: Qualifying task only after the byte cap
    # Create a first line that is 110 bytes long (exceeding 100 limit)
    filler = "x" * 100
    create_zip(
        [
            json.dumps({"filler": filler}),
            json.dumps({"tasks": [{"sequence": 1, "taskType": "LUTZ"}]}),
        ]
    )
    assert has_tasks(zip_path) is False

    # Case 3: Qualifying task early (within limits) followed by lots of filler
    create_zip(
        [
            json.dumps({"tasks": [{"sequence": 1, "taskType": "LUTZ"}]}),
            json.dumps({"msg": "lots of filler"}),
            json.dumps({"msg": "exceeding caps now..."}),
            json.dumps({"msg": "and more..."}),
        ]
    )
    assert has_tasks(zip_path) is True



from unittest.mock import AsyncMock, MagicMock

import pytest

from brokk_code.app import BrokkApp, ContextModalScreen
from brokk_code.widgets.chat_panel import ChatPanel


@pytest.mark.asyncio
async def test_slash_context_toggles_fullscreen_modal_and_syncs_token_bar_visibility():
    """
    Verify that the /context slash command opens the context modal and token bar visibility.
    """
    mock_executor = MagicMock()
    mock_executor.stop = AsyncMock()
    app = BrokkApp(executor=mock_executor)

    async with app.run_test() as pilot:
        chat_panel = app.query_one("#chat-main", ChatPanel)
        token_usage = chat_panel.query_one("#chat-token-bar")

        # Initial state
        assert not isinstance(app.screen, ContextModalScreen)
        assert not token_usage.has_class("hidden")

        # Open via slash command
        await pilot.press("/", *"context".split(), "enter")
        assert isinstance(app.screen, ContextModalScreen)
        assert token_usage.has_class("hidden")

        # Close via Escape (as slash command isn't available while modal is focused)
        await pilot.press("escape")
        assert not isinstance(app.screen, ContextModalScreen)
        assert not token_usage.has_class("hidden")



from pathlib import Path

from brokk_code.acp_server import (
    BASE_MODEL_IDS,
    DEFAULT_MODEL_SELECTION,
    DEFAULT_REASONING_LEVEL,
    REASONING_LEVEL_IDS,
    BrokkAcpBridge,
    _available_model_names,
    _build_available_models,
    _extract_session_id_for_cancel,
    _fetch_normalized_catalog_with_retries,
    _model_variants_for_model,
    _normalize_model_catalog,
    _parse_model_selection,
    _reasoning_options_for_model,
    _sanitize_reasoning_level_for_model,
    build_context_chip_blocks,
    extract_prompt_text,
    map_executor_event_to_session_update,
    normalize_mode,
    resolve_model_selection,
)


def _text_block(value: str) -> dict[str, str]:
    return {"sessionUpdate": "agent_message_chunk", "text": value}


def _thought_block(value: str) -> dict[str, str]:
    return {"sessionUpdate": "agent_thought_chunk", "text": value}


def test_normalize_mode_defaults_and_known_values() -> None:
    assert normalize_mode(None) == "LUTZ"
    assert normalize_mode("") == "LUTZ"
    assert normalize_mode("ask") == "ASK"
    assert normalize_mode("code") == "CODE"
    assert normalize_mode("lutz") == "LUTZ"
    assert normalize_mode("invalid") == "LUTZ"


def test_model_and_reasoning_constants() -> None:
    assert DEFAULT_MODEL_SELECTION in BASE_MODEL_IDS
    assert DEFAULT_REASONING_LEVEL in REASONING_LEVEL_IDS


def test_resolve_model_selection_variant_and_plain() -> None:
    assert resolve_model_selection(DEFAULT_MODEL_SELECTION) == (DEFAULT_MODEL_SELECTION, None)
    assert resolve_model_selection("gpt-5.2#r=low") == ("gpt-5.2", "low")
    assert resolve_model_selection("gemini-3-flash-preview") == ("gemini-3-flash-preview", None)


def test_normalize_model_catalog_and_reasoning_options() -> None:
    catalog = _normalize_model_catalog(
        {
            "models": [
                {
                    "name": "gpt-5.2",
                    "location": "openai/gpt-5.2",
                    "supportsReasoningEffort": True,
                    "supportsReasoningDisable": True,
                },
                {
                    "name": "gemini-3-flash-preview",
                    "location": "google/gemini-3-flash-preview",
                    "supportsReasoningEffort": False,
                    "supportsReasoningDisable": False,
                },
            ]
        }
    )
    assert [m["name"] for m in catalog] == ["gpt-5.2", "gemini-3-flash-preview"]
    assert _reasoning_options_for_model("gpt-5.2", catalog) == [
        "default",
        "low",
        "medium",
        "high",
        "disable",
    ]
    assert _reasoning_options_for_model("gemini-3-flash-preview", catalog) == [
        "default",
        "low",
        "medium",
        "high",
        "disable",
    ]
    assert (
        _sanitize_reasoning_level_for_model("gemini-3-flash-preview", "high", catalog) == "default"
    )


def test_build_available_models_includes_model_variants_with_conditional_disable() -> None:
    catalog = _normalize_model_catalog(
        {
            "models": [
                {
                    "name": "gpt-5.2",
                    "supportsReasoningEffort": True,
                    "supportsReasoningDisable": True,
                },
                {
                    "name": "gemini-3-flash-preview",
                    "supportsReasoningEffort": False,
                    "supportsReasoningDisable": False,
                },
            ]
        }
    )
    assert _model_variants_for_model("gpt-5.2", catalog) == ["low", "medium", "high", "disable"]
    assert _model_variants_for_model("gemini-3-flash-preview", catalog) == []
    options = _build_available_models(catalog)
    assert [value for value, _ in options] == [
        "gpt-5.2",
        "gpt-5.2/low",
        "gpt-5.2/medium",
        "gpt-5.2/high",
        "gpt-5.2/disable",
        "gemini-3-flash-preview",
    ]
    assert [name for _, name in options] == [
        "gpt-5.2",
        "gpt-5.2 (low)",
        "gpt-5.2 (medium)",
        "gpt-5.2 (high)",
        "gpt-5.2 (disable)",
        "gemini-3-flash-preview",
    ]
    assert all("#r=" not in value for value, _ in options)


def test_parse_model_selection_routes_model_and_variant() -> None:
    catalog = _normalize_model_catalog(
        {
            "models": [
                {
                    "name": "gpt-5.2",
                    "supportsReasoningEffort": True,
                    "supportsReasoningDisable": True,
                },
            ]
        }
    )
    assert _parse_model_selection("gpt-5.2", catalog) == ("gpt-5.2", None)
    assert _parse_model_selection("gpt-5.2/high", catalog) == ("gpt-5.2", "high")
    assert _parse_model_selection("gpt-5.2/default", catalog) == ("gpt-5.2/default", None)
    assert _parse_model_selection("model/gpt-5.2", catalog) == ("gpt-5.2", None)


def test_available_model_names_filters_invalid_entries_and_preserves_order() -> None:
    catalog = [
        {"name": "alpha"},
        {"name": None},
        {"name": "   "},
        {"name": "beta"},
        {"name": "alpha"},
        {"location": "missing-name"},
    ]

    assert _available_model_names(catalog) == ["alpha", "beta", "alpha"]


def test_parse_model_selection_ignores_none_model_name() -> None:
    catalog = [{"name": None}, {"name": "gpt-5.2"}]

    assert _parse_model_selection("None", catalog) == ("None", None)
    assert _parse_model_selection("gpt-5.2", catalog) == ("gpt-5.2", None)


async def test_fetch_normalized_catalog_with_retries_recovers_after_transient_failures() -> None:
    calls = {"count": 0}

    async def fetch_payload() -> dict[str, object]:
        calls["count"] += 1
        if calls["count"] < 3:
            raise RuntimeError("executor not ready")
        return {"models": [{"name": "gpt-5.2"}]}

    catalog = await _fetch_normalized_catalog_with_retries(
        fetch_payload,
        attempts=4,
        initial_backoff_seconds=0,
    )

    assert catalog == [
        {
            "name": "gpt-5.2",
            "location": "gpt-5.2",
            "supportsReasoningEffort": False,
            "supportsReasoningDisable": False,
        }
    ]
    assert calls["count"] == 3


async def test_fetch_normalized_catalog_with_retries_returns_none_when_catalog_stays_empty() -> (
    None
):
    calls = {"count": 0}

    async def fetch_payload() -> dict[str, object]:
        calls["count"] += 1
        return {"models": []}

    catalog = await _fetch_normalized_catalog_with_retries(
        fetch_payload,
        attempts=3,
        initial_backoff_seconds=0,
    )

    assert catalog is None
    assert calls["count"] == 3


def test_extract_prompt_text_from_blocks() -> None:
    prompt = [
        {"type": "text", "text": "Hello"},
        {"type": "image", "url": "http://example.com"},
        {"type": "text", "text": "World"},
    ]
    assert extract_prompt_text(prompt) == "Hello\nWorld"


def test_extract_prompt_text_from_string() -> None:
    assert extract_prompt_text("  direct prompt  ") == "direct prompt"


def test_map_executor_token_event() -> None:
    event = {"type": "LLM_TOKEN", "data": {"token": "abc"}}
    assert map_executor_event_to_session_update(event, _text_block) == {
        "sessionUpdate": "agent_message_chunk",
        "text": "abc",
    }


def test_map_executor_reasoning_token_event() -> None:
    event = {"type": "LLM_TOKEN", "data": {"token": "thinking", "isReasoning": True}}
    # Should use thought block when isReasoning is True
    assert map_executor_event_to_session_update(event, _text_block, _thought_block) == {
        "sessionUpdate": "agent_thought_chunk",
        "text": "thinking",
    }


def test_map_executor_reasoning_token_event_strict_parsing() -> None:
    # Truthy strings
    for val in ["true", "True", " 1 ", "yes"]:
        event = {"type": "LLM_TOKEN", "data": {"token": "t", "isReasoning": val}}
        assert map_executor_event_to_session_update(event, _text_block, _thought_block) == {
            "sessionUpdate": "agent_thought_chunk",
            "text": "t",
        }

    # Falsy strings (previously incorrectly treated as True by bool())
    for val in ["false", "0", "no", ""]:
        event = {"type": "LLM_TOKEN", "data": {"token": "t", "isReasoning": val}}
        assert map_executor_event_to_session_update(event, _text_block, _thought_block) == {
            "sessionUpdate": "agent_message_chunk",
            "text": "t",
        }

    # Booleans
    event_true = {"type": "LLM_TOKEN", "data": {"token": "t", "isReasoning": True}}
    assert map_executor_event_to_session_update(event_true, _text_block, _thought_block) == {
        "sessionUpdate": "agent_thought_chunk",
        "text": "t",
    }

    event_false = {"type": "LLM_TOKEN", "data": {"token": "t", "isReasoning": False}}
    assert map_executor_event_to_session_update(event_false, _text_block, _thought_block) == {
        "sessionUpdate": "agent_message_chunk",
        "text": "t",
    }


def test_map_executor_error_event() -> None:
    event = {"type": "ERROR", "data": {"message": "boom"}}
    assert map_executor_event_to_session_update(event, _text_block) == {
        "sessionUpdate": "agent_message_chunk",
        "text": "\n[ERROR] boom\n",
    }


def test_map_executor_unknown_event() -> None:
    event = {"type": "STATE_HINT", "data": {"name": "workspaceUpdated"}}
    assert map_executor_event_to_session_update(event, _text_block) is None


def test_map_executor_notification_event_surfaces_as_message() -> None:
    event = {"type": "NOTIFICATION", "data": {"level": "INFO", "message": "planning"}}
    assert map_executor_event_to_session_update(event, _text_block, _thought_block) == {
        "sessionUpdate": "agent_message_chunk",
        "text": "\n\n[INFO] planning\n",
    }


def test_map_executor_state_hint_surfaces_as_message() -> None:
    event = {"type": "STATE_HINT", "data": {"message": "indexing workspace"}}
    assert map_executor_event_to_session_update(event, _text_block, _thought_block) is None


def test_map_executor_warning_notification_event_surfaces_as_message() -> None:
    event = {"type": "NOTIFICATION", "data": {"level": "WARN", "message": "slow network"}}
    assert map_executor_event_to_session_update(event, _text_block, _thought_block) == {
        "sessionUpdate": "agent_message_chunk",
        "text": "\n\n[WARN] slow network\n",
    }


def test_map_executor_cost_notification_event_is_suppressed() -> None:
    event = {"type": "NOTIFICATION", "data": {"level": "COST", "message": "$0.0012 for gpt"}}
    assert map_executor_event_to_session_update(event, _text_block, _thought_block) is None


def test_map_executor_status_token_is_passed_through() -> None:
    token = "\n**Brokk** performing initial workspace review..."
    event = {"type": "LLM_TOKEN", "data": {"token": token}}
    assert map_executor_event_to_session_update(event, _text_block, _thought_block) == {
        "sessionUpdate": "agent_message_chunk",
        "text": token,
    }


def test_map_executor_status_token_mojibake_is_minimally_normalized() -> None:
    token = "\n**Brokk** performing initial workspace reviewâ€¦"
    event = {"type": "LLM_TOKEN", "data": {"token": token}}
    assert map_executor_event_to_session_update(event, _text_block, _thought_block) == {
        "sessionUpdate": "agent_message_chunk",
        "text": "\n**Brokk** performing initial workspace review...",
    }


def test_extract_session_id_for_cancel() -> None:
    assert _extract_session_id_for_cancel((), {"session_id": "abc"}) == "abc"
    assert _extract_session_id_for_cancel(({"sessionId": "def"},), {}) == "def"
    assert _extract_session_id_for_cancel((), {"params": {"sessionId": "ghi"}}) == "ghi"
    assert _extract_session_id_for_cancel((), {}) is None


async def test_ensure_ready_bootstraps_session_before_wait_ready() -> None:
    calls: list[str] = []

    class StubExecutor:
        async def start(self) -> None:
            calls.append("start")

        async def create_session(self, name: str = "ignored") -> str:
            calls.append(f"create_session:{name}")
            return "session-1"

        async def wait_ready(self) -> bool:
            calls.append("wait_ready")
            return True

    bridge = BrokkAcpBridge(StubExecutor())  # type: ignore[arg-type]
    await bridge.ensure_ready()

    assert calls == ["start", "create_session:ACP Bootstrap Session", "wait_ready"]


def test_build_context_chip_blocks() -> None:
    context_data = {
        "usedTokens": 1500,
        "maxTokens": 100000,
        "fragments": [
            {"chipKind": "EDIT", "shortDescription": "foo.py:10-20", "tokens": 400},
            {
                "chipKind": "HISTORY",
                "shortDescription": "Prior chat",
                "tokens": 250,
                "pinned": True,
            },
        ],
    }
    blocks = build_context_chip_blocks(context_data, {})
    assert blocks == []


def test_build_context_chip_blocks_empty() -> None:
    blocks = build_context_chip_blocks({"usedTokens": 10, "fragments": []}, {})
    assert blocks == []


def test_build_context_chip_blocks_uses_fragment_resource_payload() -> None:
    context_data = {
        "fragments": [
            {"id": "frag-1", "chipKind": "EDIT", "shortDescription": "script.py"},
        ]
    }
    blocks = build_context_chip_blocks(
        context_data,
        {
            "frag-1": {
                "uri": "file:///tmp/script.py",
                "mimeType": "text/x-python",
                "text": "def hello():\n    print('Hello, world!')\n",
            }
        },
    )
    assert blocks[0]["uri"] == "file:///tmp/script.py"
    assert blocks[0]["mime_type"] == "text/x-python"
    assert "def hello():" in blocks[0]["text"]


def test_build_context_chip_blocks_uses_short_description_for_brokk_fragment_uri() -> None:
    context_data = {
        "fragments": [
            {
                "id": "frag-1",
                "chipKind": "TASK_LIST",
                "shortDescription": "Add auth migration tests",
            },
        ]
    }
    blocks = build_context_chip_blocks(
        context_data,
        {
            "frag-1": {
                "uri": "brokk://context/fragment/e961b14e-927e-41c6-ae03-fd8fd0b37d8c",
                "mimeType": "text/plain",
                "text": "task list content",
            }
        },
    )
    assert blocks[0]["uri"] == "brokk://context/add-auth-migration-tests"
    assert blocks[0]["text"] == "task list content"


def test_build_context_chip_blocks_orders_by_chip_kind() -> None:
    context_data = {
        "fragments": [
            {"id": "h", "chipKind": "HISTORY"},
            {"id": "o", "chipKind": "OTHER"},
            {"id": "e", "chipKind": "EDIT"},
            {"id": "s", "chipKind": "SUMMARY"},
        ]
    }
    blocks = build_context_chip_blocks(
        context_data,
        {
            "h": {"uri": "file:///tmp/h.txt", "mimeType": "text/plain", "text": "h"},
            "o": {"uri": "file:///tmp/o.txt", "mimeType": "text/plain", "text": "o"},
            "e": {"uri": "file:///tmp/e.txt", "mimeType": "text/plain", "text": "e"},
            "s": {"uri": "file:///tmp/s.txt", "mimeType": "text/plain", "text": "s"},
        },
    )
    assert [block["uri"] for block in blocks] == [
        "file:///tmp/e.txt",
        "file:///tmp/s.txt",
        "file:///tmp/h.txt",
        "file:///tmp/o.txt",
    ]


async def test_prompt_emits_context_snapshot_after_stream(tmp_path: Path) -> None:
    updates: list[tuple[str, dict[str, str]]] = []

    class StubExecutor:
        def __init__(self, workspace_dir: Path):
            self.workspace_dir = workspace_dir

        async def start(self) -> None:
            pass

        async def create_session(self, name: str = "ignored") -> str:
            return "session-1"

        async def wait_ready(self) -> bool:
            return True

        # Accept session_id kwarg for compatibility with new API.
        async def submit_job(
            self,
            task_input: str,
            planner_model: str,
            code_model: str | None = None,
            reasoning_level: str | None = None,
            reasoning_level_code: str | None = None,
            mode: str = "LUTZ",
            session_id: str | None = None,
        ) -> str:
            # For this test the executor session should match the created session.
            assert session_id == "session-1"
            return "job-1"

        async def stream_events(self, job_id: str):
            yield {"type": "LLM_TOKEN", "data": {"token": "abc"}}

        async def get_context(self) -> dict[str, object]:
            return {
                "usedTokens": 250,
                "maxTokens": 1000,
                "fragments": [
                    {
                        "id": "frag-1",
                        "chipKind": "EDIT",
                        "shortDescription": "main.py",
                        "tokens": 50,
                    }
                ],
            }

        async def get_context_fragment(self, fragment_id: str) -> dict[str, object]:
            assert fragment_id == "frag-1"
            return {
                "id": "frag-1",
                "uri": "file:///repo/main.py",
                "mimeType": "text/x-python",
                "text": "print('hi')\n",
            }

    async def send_update(session_id: str, update: dict[str, str]) -> None:
        updates.append((session_id, update))

    def update_agent_message_text(text: str) -> dict[str, str]:
        return {"sessionUpdate": "agent_message_chunk", "text": text}

    def update_agent_thought_text(text: str) -> dict[str, str]:
        return {"sessionUpdate": "agent_thought_chunk", "text": text}

    def build_context_snapshot_update(uri: str, mime_type: str, text: str) -> dict[str, str]:
        return {
            "sessionUpdate": "agent_message_chunk",
            "uri": uri,
            "mimeType": mime_type,
            "text": text,
            "kind": "embedded_resource",
        }

    bridge = BrokkAcpBridge(StubExecutor(tmp_path))  # type: ignore[arg-type]
    await bridge.prompt(
        prompt=[{"type": "text", "text": "hello"}],
        session_id="acp-session-1",
        mode="LUTZ",
        planner_model="gpt-5.2",
        code_model="gemini-3-flash-preview",
        reasoning_level="low",
        reasoning_level_code="disable",
        send_update=send_update,
        update_agent_message_text=update_agent_message_text,
        update_agent_thought_text=update_agent_thought_text,
        build_context_snapshot_update=build_context_snapshot_update,
    )

    assert len(updates) == 7
    assert updates[0][1]["text"] == "abc"
    assert "Context Snapshot" in updates[1][1]["text"]
    assert "data:image/svg+xml;base64," in updates[2][1]["text"]
    assert "Editable Context" in updates[3][1]["text"]
    assert updates[4][1]["text"] == "- "
    assert updates[5][1]["kind"] == "embedded_resource"
    assert updates[5][1]["uri"] == "file:///repo/main.py"
    assert updates[5][1]["mimeType"] == "text/x-python"
    assert updates[5][1]["text"] == "print('hi')\n"
    assert updates[6][1]["text"] == " | 50\n"


async def test_prompt_emits_discarded_context_as_json_markdown_text(tmp_path: Path) -> None:
    updates: list[tuple[str, dict[str, str]]] = []

    class StubExecutor:
        def __init__(self, workspace_dir: Path):
            self.workspace_dir = workspace_dir

        async def start(self) -> None:
            pass

        async def create_session(self, name: str = "ignored") -> str:
            return "session-1"

        async def wait_ready(self) -> bool:
            return True

        async def submit_job(
            self,
            task_input: str,
            planner_model: str,
            code_model: str | None = None,
            reasoning_level: str | None = None,
            reasoning_level_code: str | None = None,
            mode: str = "LUTZ",
            session_id: str | None = None,
        ) -> str:
            assert session_id == "session-1"
            return "job-1"

        async def stream_events(self, job_id: str):
            yield {"type": "LLM_TOKEN", "data": {"token": "abc"}}

        async def get_context(self) -> dict[str, object]:
            return {
                "usedTokens": 250,
                "maxTokens": 1000,
                "fragments": [
                    {
                        "id": "frag-dc",
                        "chipKind": "OTHER",
                        "shortDescription": "Discarded Context",
                        "tokens": 10,
                    }
                ],
            }

        async def get_context_fragment(self, fragment_id: str) -> dict[str, object]:
            assert fragment_id == "frag-dc"
            return {
                "id": "frag-dc",
                "uri": "brokk://context/fragment/frag-dc",
                "mimeType": "text/plain",
                "text": "dropped items here",
            }

    async def send_update(session_id: str, update: dict[str, str]) -> None:
        updates.append((session_id, update))

    def update_agent_message_text(text: str) -> dict[str, str]:
        return {"sessionUpdate": "agent_message_chunk", "text": text}

    def update_agent_thought_text(text: str) -> dict[str, str]:
        return {"sessionUpdate": "agent_thought_chunk", "text": text}

    def build_context_snapshot_update(uri: str, mime_type: str, text: str) -> dict[str, str]:
        return {
            "sessionUpdate": "agent_message_chunk",
            "uri": uri,
            "mimeType": mime_type,
            "text": text,
            "kind": "embedded_resource",
        }

    bridge = BrokkAcpBridge(StubExecutor(tmp_path))  # type: ignore[arg-type]
    await bridge.prompt(
        prompt=[{"type": "text", "text": "hello"}],
        session_id="acp-session-1",
        mode="LUTZ",
        planner_model="gpt-5.2",
        code_model="gemini-3-flash-preview",
        reasoning_level="low",
        reasoning_level_code="disable",
        send_update=send_update,
        update_agent_message_text=update_agent_message_text,
        update_agent_thought_text=update_agent_thought_text,
        build_context_snapshot_update=build_context_snapshot_update,
    )

    json_blocks = [u[1]["text"] for u in updates if "```json" in u[1].get("text", "")]
    assert len(json_blocks) == 1
    assert '"title": "Discarded Context"' in json_blocks[0]
    assert '"content": "dropped items here"' in json_blocks[0]


async def test_prompt_emits_summary_as_list_item_with_text_and_tokens_for_brokk_uri(
    tmp_path: Path,
) -> None:
    updates: list[tuple[str, dict[str, str]]] = []

    class StubExecutor:
        def __init__(self, workspace_dir: Path):
            self.workspace_dir = workspace_dir

        async def start(self) -> None:
            pass

        async def create_session(self, name: str = "ignored") -> str:
            return "session-1"

        async def wait_ready(self) -> bool:
            return True

        async def submit_job(
            self,
            task_input: str,
            planner_model: str,
            code_model: str | None = None,
            reasoning_level: str | None = None,
            reasoning_level_code: str | None = None,
            mode: str = "LUTZ",
            session_id: str | None = None,
        ) -> str:
            assert session_id == "session-1"
            return "job-1"

        async def stream_events(self, job_id: str):
            yield {"type": "LLM_TOKEN", "data": {"token": "abc"}}

        async def get_context(self) -> dict[str, object]:
            return {
                "usedTokens": 250,
                "maxTokens": 1000,
                "fragments": [
                    {
                        "id": "frag-s",
                        "chipKind": "SUMMARY",
                        "shortDescription": "Summary of worker.go",
                        "tokens": 12,
                    }
                ],
            }

        async def get_context_fragment(self, fragment_id: str) -> dict[str, object]:
            assert fragment_id == "frag-s"
            return {
                "id": "frag-s",
                "uri": "brokk://context/fragment/frag-s",
                "mimeType": "text/plain",
                "text": "summary text",
            }

    async def send_update(session_id: str, update: dict[str, str]) -> None:
        updates.append((session_id, update))

    def update_agent_message_text(text: str) -> dict[str, str]:
        return {"sessionUpdate": "agent_message_chunk", "text": text}

    def update_agent_thought_text(text: str) -> dict[str, str]:
        return {"sessionUpdate": "agent_thought_chunk", "text": text}

    def build_context_snapshot_update(uri: str, mime_type: str, text: str) -> dict[str, str]:
        return {
            "sessionUpdate": "agent_message_chunk",
            "uri": uri,
            "mimeType": mime_type,
            "text": text,
            "kind": "embedded_resource",
        }

    bridge = BrokkAcpBridge(StubExecutor(tmp_path))  # type: ignore[arg-type]
    await bridge.prompt(
        prompt=[{"type": "text", "text": "hello"}],
        session_id="acp-session-1",
        mode="LUTZ",
        planner_model="gpt-5.2",
        code_model="gemini-3-flash-preview",
        reasoning_level="low",
        reasoning_level_code="disable",
        send_update=send_update,
        update_agent_message_text=update_agent_message_text,
        update_agent_thought_text=update_agent_thought_text,
        build_context_snapshot_update=build_context_snapshot_update,
    )

    assert len(updates) == 7
    assert updates[0][1]["text"] == "abc"
    assert "Context Snapshot" in updates[1][1]["text"]
    assert "data:image/svg+xml;base64," in updates[2][1]["text"]
    assert "Summaries" in updates[3][1]["text"]
    assert updates[4][1]["text"] == "- "
    assert updates[5][1]["text"] == "Summary of worker.go"
    assert updates[6][1]["text"] == " | 12\n"



from pathlib import Path


def resolve_workspace_dir(path: Path) -> Path:
    """Resolve a workspace path to its git repository root when applicable."""
    resolved = path.resolve()
    current = resolved if resolved.is_dir() else resolved.parent

    for candidate in (current, *current.parents):
        git_path = candidate / ".git"
        if git_path.is_dir() or git_path.is_file():
            return candidate

    return resolved



from unittest.mock import AsyncMock, MagicMock

import pytest
from rich.text import Text
from textual.widgets import Static

from brokk_code.widgets.chat_panel import ChatPanel


@pytest.mark.asyncio
async def test_token_usage_update():
    """Verify that updating token usage updates the widget text."""
    from textual.app import App, ComposeResult

    from brokk_code.widgets.token_bar import TokenBar

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test():
        panel = app.query_one("#chat", ChatPanel)
        token_bar = panel.query_one("#chat-token-bar", TokenBar)

        # Initial empty - should show no context message
        assert "No context yet" in str(token_bar.render())

        # Update with used and max
        panel.set_token_usage(1500, 100000)
        assert "98.5% context remaining" in str(token_bar.render())

        # Update with half usage
        panel.set_token_usage(50000, 100000)
        assert "50.0% context remaining" in str(token_bar.render())

        # Update with only used
        panel.set_token_usage(2500)
        # TokenBar defaults max to 200,000 if not provided
        # 1 - (2500/200000) = 0.9875 -> 98.8%
        assert "98.8% context remaining" in str(token_bar.render())

        # Update with 0 clears it
        panel.set_token_usage(0)
        assert "No context yet" in str(token_bar.render())


@pytest.mark.asyncio
async def test_job_progress_in_chat_panel():
    """
    Verify that job running state is reflected in ChatPanel's status timer
    and the help row spinner.
    """
    from textual.app import App, ComposeResult

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test():
        chat = app.query_one(ChatPanel)
        help_elapsed = chat.query_one("#help-elapsed")
        help_spinner = chat.query_one("#help-spinner")

        # Initially hidden
        assert help_elapsed.has_class("hidden")
        assert help_spinner.has_class("hidden")

        # Start job
        chat.set_job_running(True)
        assert not help_elapsed.has_class("hidden")
        assert not help_spinner.has_class("hidden")

        # Stop job
        chat.set_job_running(False)
        assert help_elapsed.has_class("hidden")
        assert help_spinner.has_class("hidden")


@pytest.mark.asyncio
async def test_token_bar_visibility_control():
    """Verify that the token bar visibility can be controlled explicitly."""
    from textual.app import App, ComposeResult

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test():
        panel = app.query_one("#chat", ChatPanel)
        token_bar = panel.query_one("#chat-token-bar")

        # Initial state: hidden
        assert token_bar.has_class("hidden")

        # Set visible
        panel.set_token_bar_visible(True)
        assert not token_bar.has_class("hidden")

        # Set hidden
        panel.set_token_bar_visible(False)
        assert token_bar.has_class("hidden")


@pytest.mark.asyncio
async def test_streaming_duplication_regression():
    """
    Verify that streaming tokens incrementally does not result in duplicated
    content in the RichLog.
    """
    from textual.app import App, ComposeResult
    from textual.widgets import RichLog

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test() as pilot:
        panel = app.query_one("#chat", ChatPanel)
        log = panel.query_one("#chat-log", RichLog)

        # Simulate a stream: "Hello" -> "Hello world" -> "Hello world!"
        # We use a short flush interval to ensure incremental flushes trigger.
        panel._flush_interval = 0.01

        panel.append_token(
            "Hello", "AI", is_new_message=True, is_reasoning=False, is_terminal=False
        )
        await pilot.pause()

        panel.append_token(
            " world", "AI", is_new_message=False, is_reasoning=False, is_terminal=False
        )
        await pilot.pause()

        panel.append_token("!", "AI", is_new_message=False, is_reasoning=False, is_terminal=True)
        await pilot.pause()

        # Check the rendered lines in the log.
        # We expect "Hello world!" to appear exactly once.
        # It should not appear as:
        # Hello
        # Hello world
        # Hello world!

        # log.lines contains the objects passed to write()
        content_strings = [str(line) for line in log.lines]
        combined_text = "".join(content_strings)

        # Count occurrences of the final string
        count = combined_text.count("Hello world!")
        assert count == 1, (
            f"Expected 'Hello world!' once, found {count}. Content: {content_strings}"
        )


@pytest.mark.asyncio
async def test_whitespace_reasoning_terminal_does_not_stick():
    """
    Regression test: If a reasoning token that is only whitespace is flushed as terminal,
    the panel must not remain in reasoning mode for the next non-reasoning message.
    """
    from textual.app import App, ComposeResult
    from textual.widgets import RichLog

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test() as pilot:
        panel = app.query_one("#chat", ChatPanel)
        log = panel.query_one("#chat-log", RichLog)

        # 1) Simulate a reasoning stream that only emits whitespace and is terminal.
        panel.append_token("   ", "AI", is_new_message=True, is_reasoning=True, is_terminal=True)
        await pilot.pause()

        # Ensure nothing meaningful was rendered as a Thinking panel
        combined = "".join(str(line) for line in log.lines)
        assert "Thinking" not in combined

        # 2) Now append a normal non-reasoning message and ensure it renders as Markdown
        panel.append_token("Hello", "AI", is_new_message=True, is_reasoning=False, is_terminal=True)
        await pilot.pause()

        combined = "".join(str(line) for line in log.lines)
        # Should contain the Markdown-rendered Hello, and still not contain a Thinking panel.
        assert "Hello" in combined
        assert "Thinking" not in combined


@pytest.mark.asyncio
async def test_no_notification_panel_widget():
    """Verify that ChatPanel does not contain a notification panel widget."""
    from textual.app import App, ComposeResult
    from textual.css.query import NoMatches

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test():
        chat = app.query_one("#chat", ChatPanel)
        # It should not have a widget with id notification-panel
        with pytest.raises(NoMatches):
            chat.query_one("#notification-panel")


@pytest.mark.asyncio
async def test_notification_routing_to_chat_log():
    """Verify that notifications are routed to the main chat log with styling."""
    from textual.app import App, ComposeResult
    from textual.widgets import RichLog

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test() as pilot:
        panel = app.query_one("#chat", ChatPanel)
        log = panel.query_one("#chat-log", RichLog)

        # Test INFO notification (no prefix)
        panel.add_system_message("Info message", level="INFO")
        await pilot.pause()
        # Ensure 'Info message' is present but NOT explicitly prefixed with '[INFO]'
        assert any(
            "Info message" in str(line) and not str(line).startswith("[INFO]") for line in log.lines
        )

        # Test ERROR notification (prefixed)
        panel.add_system_message("Error message", level="ERROR")
        await pilot.pause()
        assert any("[ERROR] Error message" in str(line) for line in log.lines)

        # Test COST notification (prefixed)
        panel.add_system_message("Cost message", level="COST")
        await pilot.pause()
        assert any("[COST] Cost message" in str(line) for line in log.lines)


@pytest.mark.asyncio
async def test_action_handle_ctrl_c_no_input_widget():
    """
    Regression test: Verify action_handle_ctrl_c does not crash if _maybe_chat()
    returns a panel but the #chat-input widget is not currently in the DOM.
    """
    from unittest.mock import AsyncMock, MagicMock, patch

    from brokk_code.app import BrokkApp

    executor = MagicMock()
    executor.stop = AsyncMock()
    app = BrokkApp(executor=executor)

    with (
        patch.object(BrokkApp, "_start_executor", return_value=None),
        patch.object(BrokkApp, "_monitor_executor", return_value=None),
        patch.object(BrokkApp, "_poll_tasklist", return_value=None),
        patch.object(BrokkApp, "_poll_context", return_value=None),
    ):
        async with app.run_test() as pilot:
            # Simulate a job in progress so we can verify the fall-through behavior
            app.job_in_progress = True
            app.current_job_id = "test-job-id"
            app.executor.cancel_job = AsyncMock()

            # Mock query to return empty when looking for #chat-input
            # This simulates the widget being unmounted/missing
            original_query = app.query

            def mocked_query(selector: str):
                if selector == "#chat-input":
                    empty_result = MagicMock()
                    empty_result.results.return_value = iter([])
                    return empty_result
                return original_query(selector)

            with patch.object(app, "query", side_effect=mocked_query):
                # This should NOT raise even though #chat-input is 'missing'
                await app.action_handle_ctrl_c()
                await pilot.pause()

            # Verify it fell through to job cancellation
            app.executor.cancel_job.assert_called_once_with("test-job-id")


@pytest.mark.asyncio
async def test_no_ctrl_u_e_bindings_in_chat_input():
    """Verify that ChatInput does not have ctrl+u or ctrl+e bindings."""
    from brokk_code.widgets.chat_panel import ChatInput

    bindings = {b.key for b in ChatInput.BINDINGS}
    assert "ctrl+u" not in bindings
    assert "ctrl+e" not in bindings


def _static_rendered_text(widget: Static) -> str:
    rendered = widget.render()
    if isinstance(rendered, Text):
        return rendered.plain
    return str(rendered)


@pytest.mark.asyncio
async def test_chat_panel_composition_success():
    """
    Verify that ChatPanel composes without NameError and contains expected widgets.
    Regression test for missing Static import.
    """
    from textual.app import App, ComposeResult
    from textual.widgets import Static

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test():
        # Verify ChatPanel exists
        chat = app.query_one("#chat", ChatPanel)
        assert chat is not None

        # Verify #help-elapsed (Static) exists - this would have failed with NameError
        help_elapsed = chat.query_one("#help-elapsed", Static)
        assert help_elapsed is not None


@pytest.mark.asyncio
async def test_chat_help_line_includes_shift_tab_mode_after_history() -> None:
    from textual.app import App, ComposeResult

    class TestApp(App):
        def compose(self) -> ComposeResult:
            yield ChatPanel(id="chat")

    app = TestApp()
    async with app.run_test():
        chat = app.query_one("#chat", ChatPanel)
        help_line = chat.query_one("#chat-help", Static)
        rendered = _static_rendered_text(help_line)

        assert "Up/Down: History" in rendered
        assert "Shift+Tab: Mode" in rendered
        assert "/commands" not in rendered
        assert rendered.index("Up/Down: History") < rendered.index("Shift+Tab: Mode")


@pytest.mark.asyncio
async def test_slash_command_catalog_stability():
    """Verify the slash command catalog is stable and follows rules."""
    from unittest.mock import MagicMock

    from brokk_code.app import BrokkApp

    app = BrokkApp(executor=MagicMock())
    commands = app.get_slash_commands()

    assert len(commands) > 0
    seen = set()
    for cmd_entry in commands:
        cmd = cmd_entry["command"]
        assert cmd.startswith("/"), f"Command {cmd} must start with /"
        assert cmd not in seen, f"Duplicate command found: {cmd}"
        assert "description" in cmd_entry
        seen.add(cmd)

    # Verify key commands exist
    cmds_only = {c["command"] for c in commands}
    assert "/ask" in cmds_only
    assert "/task" in cmds_only
    assert "/help" in cmds_only


@pytest.mark.asyncio
async def test_slash_autocomplete_filtering():
    """Verify slash suggestions filter correctly."""
    from textual.app import App, ComposeResult

    from brokk_code.widgets.chat_panel import ChatPanel, SlashCommandSuggestions

    class TestApp(App):
        def get_slash_commands(self):
            return [
                {"command": "/ask", "description": "d"},
                {"command": "/ask-more", "description": "d"},
                {"command": "/help", "description": "d"},
            ]

        def compose(self) -> ComposeResult:
            yield ChatPanel()

    app = TestApp()
    async with app.run_test() as pilot:
        suggestions = app.query_one(SlashCommandSuggestions)

        # Type /a
        await pilot.press(*list("/a"))
        assert suggestions.display is True
        assert len(suggestions.children) == 2

        # Type sk-
        await pilot.press(*list("sk-"))
        assert len(suggestions.children) == 1
        assert "/ask-more" in str(suggestions.children[0].query_one(Static).render())

        # Esc hides
        await pilot.press("escape")
        assert suggestions.display is False


@pytest.mark.asyncio
async def test_mention_autocomplete_fetch_and_accept():
    """Verify @mention autocomplete fetches remote completions and inserts selection."""
    from textual.app import App, ComposeResult

    from brokk_code.widgets.chat_panel import ChatInput, MentionSuggestions

    class TestApp(App):
        def __init__(self) -> None:
            super().__init__()
            self.executor = MagicMock()
            self.executor.get_completions = AsyncMock(
                return_value={
                    "completions": [
                        {
                            "type": "class",
                            "name": "ContextManager",
                            "detail": "ai.brokk.ContextManager",
                        }
                    ]
                }
            )

        def get_slash_commands(self):
            return []

        def compose(self) -> ComposeResult:
            yield ChatPanel()

    app = TestApp()
    async with app.run_test() as pilot:
        mentions = app.query_one(MentionSuggestions)
        chat_input = app.query_one("#chat-input", ChatInput)

        await pilot.press("@", "C", "o", "n")
        await pilot.pause(0.25)

        assert mentions.display is True
        assert len(mentions.children) == 1
        app.executor.get_completions.assert_awaited()

        await pilot.press("enter")
        await pilot.pause()
        assert mentions.display is False
        assert chat_input.text == "@ai.brokk.ContextManager "


@pytest.mark.asyncio
async def test_mention_autocomplete_ignores_email_like_text():
    """Verify email-like text does not trigger @mention autocomplete."""
    from textual.app import App, ComposeResult

    from brokk_code.widgets.chat_panel import MentionSuggestions

    class TestApp(App):
        def __init__(self) -> None:
            super().__init__()
            self.executor = MagicMock()
            self.executor.get_completions = AsyncMock(return_value={"completions": []})

        def compose(self) -> ComposeResult:
            yield ChatPanel()

    app = TestApp()
    async with app.run_test() as pilot:
        mentions = app.query_one(MentionSuggestions)
        await pilot.press(*list("foo@bar"))
        await pilot.pause(0.25)

        assert mentions.display is False
        app.executor.get_completions.assert_not_awaited()



from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from rich.text import Text
from textual.widgets import Static

from brokk_code.app import BrokkApp
from brokk_code.widgets.context_panel import ContextPanel


def _static_rendered_text(widget: Static) -> str:
    rendered = widget.render()
    if isinstance(rendered, Text):
        return rendered.plain
    return str(rendered)


@pytest.mark.asyncio
async def test_context_panel_shows_clear_selection_state():
    mock_executor = MagicMock()
    mock_executor.stop = AsyncMock()
    mock_executor.get_context = AsyncMock(
        return_value={
            "usedTokens": 321,
            "maxTokens": 100000,
            "fragments": [
                {
                    "id": "f-1",
                    "chipKind": "EDIT",
                    "shortDescription": "Editable file",
                    "pinned": True,
                    "readonly": False,
                    "editable": True,
                    "tokens": 100,
                },
                {
                    "id": "f-2",
                    "chipKind": "HISTORY",
                    "shortDescription": "Conversation history",
                    "pinned": False,
                    "readonly": False,
                    "editable": False,
                    "tokens": 221,
                },
            ],
        }
    )
    app = BrokkApp(executor=mock_executor)

    with (
        patch.object(BrokkApp, "_start_executor", return_value=None),
        patch.object(BrokkApp, "_monitor_executor", return_value=None),
        patch.object(BrokkApp, "_poll_tasklist", return_value=None),
        patch.object(BrokkApp, "_poll_context", return_value=None),
    ):
        async with app.run_test() as pilot:
            app._executor_ready = True

            # Open context modal via slash command
            await pilot.press("/", *"context".split(), "enter")
            await app._refresh_context_panel()
            await pilot.pause()

            panel = app.screen.query_one("#context-panel", ContextPanel)
            assert len(panel.query(".context-chip")) >= 2

            # Verify help line is present and contains some expected keys
            help_line = panel.query_one("#context-help-line", Static)
            help_text = _static_rendered_text(help_line)

            # Esc is now first and highlighted; assert content + ordering without
            # overfitting spacing/alignment.
            assert "Esc Close" in help_text
            assert help_text.index("Esc") < help_text.index("Space")
            assert help_text.index("Esc") < help_text.index("Enter")

            # Basic keys from _get_shortcuts_text()
            assert "Space" in help_text
            assert "Enter" in help_text
            assert "Drop" in help_text
            assert "U" in help_text  # clear_selection
            assert "CTRL+A" in help_text  # select_all
            assert "P" in help_text  # toggle_pin_selected

            selection_status = panel.query_one("#context-selection-status")
            active_status = panel.query_one("#context-active-status")
            assert "Selected: 0" in str(selection_status.render())
            assert "Active: Editable file" in str(active_status.render())

            # Select the cursor chip (first fragment).
            await pilot.press("enter")
            await pilot.pause()

            assert "Selected: 1" in str(selection_status.render())
            selected_chip_texts = " ".join(
                item.render().plain for item in panel.query(".context-chip")
            )
            assert "[SELECTED]" in selected_chip_texts
            assert "[ACTIVE]" not in selected_chip_texts

            # Move active cursor to next chip.
            await pilot.press("right")
            await pilot.pause()
            assert "Active: Conversation history" in str(active_status.render())


@pytest.mark.asyncio
async def test_context_panel_drop_others_action():
    mock_executor = MagicMock()
    mock_executor.stop = AsyncMock()
    mock_executor.drop_context_fragments = AsyncMock()
    mock_executor.get_context = AsyncMock(
        return_value={
            "usedTokens": 500,
            "maxTokens": 100000,
            "fragments": [
                {
                    "id": "f-1",
                    "chipKind": "EDIT",
                    "shortDescription": "Keep Me (Active)",
                    "tokens": 100,
                },
                {
                    "id": "f-2",
                    "chipKind": "EDIT",
                    "shortDescription": "Drop Me",
                    "tokens": 100,
                },
                {
                    "id": "f-3",
                    "chipKind": "HISTORY",
                    "shortDescription": "Keep History",
                    "tokens": 100,
                },
                {
                    "id": "f-4",
                    "chipKind": "EDIT",
                    "shortDescription": "Keep Pinned",
                    "pinned": True,
                    "tokens": 100,
                },
                {
                    "id": "f-5",
                    "chipKind": "EDIT",
                    "shortDescription": "Keep Selected",
                    "tokens": 100,
                },
            ],
        }
    )
    app = BrokkApp(executor=mock_executor)

    with (
        patch.object(BrokkApp, "_start_executor", return_value=None),
        patch.object(BrokkApp, "_monitor_executor", return_value=None),
        patch.object(BrokkApp, "_poll_tasklist", return_value=None),
        patch.object(BrokkApp, "_poll_context", return_value=None),
    ):
        async with app.run_test() as pilot:
            app._executor_ready = True
            await pilot.press("/", *"context".split(), "enter")
            await app._refresh_context_panel()
            await pilot.pause()

            # f-1 is active by default. Select f-5 as well.
            await pilot.press("right", "right", "right", "right", "space")
            # Move cursor back to f-1 to make it active again.
            await pilot.press("left", "left", "left", "left")
            await pilot.pause()

            # Trigger 'drop others' (key 'o').
            # Protected: f-1 (active), f-3 (history), f-4 (pinned), f-5 (selected).
            # Dropped: f-2.
            await pilot.press("o")
            await pilot.pause()

            mock_executor.drop_context_fragments.assert_called_once_with(["f-2"])



from unittest.mock import AsyncMock, MagicMock

import pytest

from brokk_code.app import BrokkApp
from brokk_code.settings import Settings
from brokk_code.widgets.chat_panel import ChatPanel


def test_handle_command_autocommit_shows_status(tmp_path):
    app = BrokkApp(executor=MagicMock(workspace_dir=tmp_path))
    mock_chat = MagicMock(spec=ChatPanel)
    app.query_one = MagicMock(return_value=mock_chat)

    app.auto_commit = True
    app._handle_command("/autocommit")

    args, kwargs = mock_chat.add_system_message_markup.call_args
    assert "Auto-commit mode" in args[0]
    assert "ON" in args[0]
    assert kwargs.get("level") == "WARNING"


def test_handle_command_autocommit_off_persists_and_announces(tmp_path):
    app = BrokkApp(executor=MagicMock(workspace_dir=tmp_path))
    mock_chat = MagicMock(spec=ChatPanel)
    app.query_one = MagicMock(return_value=mock_chat)

    app.settings = Settings()
    app.settings.save = MagicMock()

    app.auto_commit = True
    app._handle_command("/autocommit off")

    assert app.auto_commit is False
    assert app.settings.last_auto_commit is False
    app.settings.save.assert_called_once()

    args, kwargs = mock_chat.add_system_message_markup.call_args
    assert "Auto-commit mode" in args[0]
    assert "OFF" in args[0]
    assert kwargs.get("level") == "WARNING"


@pytest.mark.asyncio
async def test_run_job_passes_auto_commit_flag(tmp_path):
    executor = MagicMock()
    executor.workspace_dir = tmp_path
    executor.submit_job = AsyncMock(return_value="job-1")

    async def stream_events(_job_id: str):
        if False:  # pragma: no cover
            yield {}

    executor.stream_events = stream_events

    app = BrokkApp(executor=executor)
    mock_chat = MagicMock(spec=ChatPanel)
    app.query_one = MagicMock(return_value=mock_chat)

    app.auto_commit = False
    await app._run_job("hello")

    assert executor.submit_job.await_count == 1
    assert executor.submit_job.await_args.kwargs["auto_commit"] is False


@pytest.mark.asyncio
async def test_run_job_attaches_at_mentions_to_context(tmp_path):
    executor = MagicMock()
    executor.workspace_dir = tmp_path
    executor.submit_job = AsyncMock(return_value="job-1")
    executor.get_completions = AsyncMock(
        side_effect=[
            {
                "completions": [
                    {
                        "type": "file",
                        "name": "app.py",
                        "detail": "brokk_code/app.py",
                    }
                ]
            },
            {
                "completions": [
                    {
                        "type": "class",
                        "name": "ContextManager",
                        "detail": "ai.brokk.ContextManager",
                    }
                ]
            },
            {
                "completions": [
                    {
                        "type": "function",
                        "name": "findUser",
                        "detail": "com.example.UserService.findUser",
                    }
                ]
            },
        ]
    )
    executor.add_context_files = AsyncMock(return_value={"added": []})
    executor.add_context_classes = AsyncMock(return_value={"added": []})
    executor.add_context_methods = AsyncMock(return_value={"added": []})

    async def stream_events(_job_id: str):
        if False:  # pragma: no cover
            yield {}

    executor.stream_events = stream_events

    app = BrokkApp(executor=executor)
    mock_chat = MagicMock(spec=ChatPanel)
    app.query_one = MagicMock(return_value=mock_chat)

    await app._run_job(
        "Use @brokk_code/app.py and @ai.brokk.ContextManager and @com.example.UserService.findUser"
    )

    executor.add_context_files.assert_awaited_once_with(["brokk_code/app.py"])
    executor.add_context_classes.assert_awaited_once_with(["ai.brokk.ContextManager"])
    executor.add_context_methods.assert_awaited_once_with(["com.example.UserService.findUser"])
    executor.submit_job.assert_awaited_once()


@pytest.mark.asyncio
async def test_run_job_skips_non_matching_at_mentions(tmp_path):
    executor = MagicMock()
    executor.workspace_dir = tmp_path
    executor.submit_job = AsyncMock(return_value="job-1")
    executor.get_completions = AsyncMock(
        return_value={
            "completions": [
                {
                    "type": "class",
                    "name": "ContextManager",
                    "detail": "ai.brokk.ContextManager",
                }
            ]
        }
    )
    executor.add_context_files = AsyncMock(return_value={"added": []})
    executor.add_context_classes = AsyncMock(return_value={"added": []})
    executor.add_context_methods = AsyncMock(return_value={"added": []})

    async def stream_events(_job_id: str):
        if False:  # pragma: no cover
            yield {}

    executor.stream_events = stream_events

    app = BrokkApp(executor=executor)
    mock_chat = MagicMock(spec=ChatPanel)
    app.query_one = MagicMock(return_value=mock_chat)

    await app._run_job("Use @DoesNotMatch")

    executor.add_context_files.assert_not_awaited()
    executor.add_context_classes.assert_not_awaited()
    executor.add_context_methods.assert_not_awaited()

    mock_chat.add_system_message.assert_any_call("No exact match for @DoesNotMatch")
    executor.submit_job.assert_awaited_once()


@pytest.mark.asyncio
async def test_run_job_rolls_back_attached_mention_fragments_on_submit_failure(tmp_path):
    executor = MagicMock()
    executor.workspace_dir = tmp_path
    executor.get_completions = AsyncMock(
        return_value={
            "completions": [
                {
                    "type": "file",
                    "name": "app.py",
                    "detail": "brokk_code/app.py",
                }
            ]
        }
    )
    executor.add_context_files = AsyncMock(return_value={"added": [{"id": "frag-1"}]})
    executor.drop_context_fragments = AsyncMock(return_value={"dropped": ["frag-1"]})
    executor.submit_job = AsyncMock(side_effect=RuntimeError("submit failed"))

    app = BrokkApp(executor=executor)
    mock_chat = MagicMock(spec=ChatPanel)
    app.query_one = MagicMock(return_value=mock_chat)

    await app._run_job("Use @brokk_code/app.py")

    executor.add_context_files.assert_awaited_once_with(["brokk_code/app.py"])
    executor.drop_context_fragments.assert_awaited_once_with(["frag-1"])
    assert executor.submit_job.await_count == 1



import argparse
import asyncio
import sys
from pathlib import Path

from brokk_code.intellij_config import configure_intellij_acp_settings
from brokk_code.workspace import resolve_workspace_dir
from brokk_code.zed_config import ExistingBrokkCodeEntryError, configure_zed_acp_settings


def _add_common_runtime_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--workspace",
        type=str,
        default=".",
        help="Path to the workspace directory (default: current directory)",
    )
    parser.add_argument(
        "--vendor",
        type=str,
        choices=["Default", "Anthropic", "Gemini", "OpenAI", "OpenAI - Codex"],
        default=None,
        help=(
            "Set 'Other Models' vendor preference (affects "
            + "internal roles like summarize/scan/commit). "
            "Use 'Default' to clear overrides."
        ),
    )
    parser.add_argument(
        "--jar",
        type=str,
        default=None,
        help="Path to brokk.jar (bypasses jbang; default: use jbang to launch)",
    )
    parser.add_argument(
        "--executor-version",
        type=str,
        default=None,
        help="[Ignored] Use jbang to manage versions (default: latest snapshot)",
    )
    parser.add_argument(
        "--executor-snapshot",
        action="store_true",
        default=True,
        help="[Ignored] Use jbang to manage versions",
    )
    parser.add_argument(
        "--executor-stable",
        action="store_false",
        dest="executor_snapshot",
        help="[Ignored] Use jbang to manage versions",
    )


def _build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Brokk Code - Interactive Terminal Interface")
    _add_common_runtime_args(parser)
    parser.add_argument(
        "--session",
        type=str,
        default=None,
        help="Attempt to resume a specific session by ID",
    )
    parser.add_argument(
        "--resume",
        action="store_true",
        dest="resume_session",
        default=False,
        help="Resume the last used session instead of creating a new one",
    )

    subparsers = parser.add_subparsers(dest="command")

    resume_parser = subparsers.add_parser("resume", help="Resume a specific session")
    _add_common_runtime_args(resume_parser)
    resume_parser.add_argument(
        "session_id",
        type=str,
        help="The ID of the session to resume",
    )

    acp_parser = subparsers.add_parser("acp", help="Run in ACP server mode")
    _add_common_runtime_args(acp_parser)
    acp_parser.add_argument(
        "--ide",
        type=str,
        choices=["intellij", "zed"],
        default="intellij",
        help="ACP client profile to target (default: intellij)",
    )

    install_parser = subparsers.add_parser("install", help="Install integration settings")
    install_parser.add_argument(
        "target",
        choices=["zed", "intellij"],
        help="Install target for integration settings",
    )
    install_parser.add_argument(
        "--force",
        action="store_true",
        default=False,
        help="Overwrite existing install configuration when supported",
    )

    return parser


def main():
    parser = _build_parser()
    args = parser.parse_args()

    if args.command == "install":
        try:
            if args.target == "zed":
                settings_path = configure_zed_acp_settings(force=args.force)
                target_name = "Zed"
            elif args.target == "intellij":
                settings_path = configure_intellij_acp_settings(force=args.force)
                target_name = "IntelliJ"
            else:
                # Should not happen due to argparse choices
                raise ValueError(f"Unknown target: {args.target}")
        except (ExistingBrokkCodeEntryError, ValueError) as exc:
            print(f"Error: {exc}", file=sys.stderr)
            sys.exit(1)

        print(f"Configured {target_name} ACP integration in {settings_path}")
        return

    workspace_path = Path(args.workspace).resolve()
    if not workspace_path.exists():
        print(f"Error: Workspace path does not exist: {workspace_path}")
        sys.exit(1)
    workspace_path = resolve_workspace_dir(workspace_path)
    jar_path = Path(args.jar).resolve() if args.jar else None

    if args.command == "acp":
        try:
            from brokk_code.acp_server import run_acp_server
        except ImportError:
            print("Error: Could not import ACP server module.", file=sys.stderr)
            sys.exit(1)

        asyncio.run(
            run_acp_server(
                workspace_dir=workspace_path,
                jar_path=jar_path,
                executor_version=args.executor_version,
                executor_snapshot=args.executor_snapshot,
                ide=args.ide,
                vendor=args.vendor,
            )
        )
        return

    try:
        from brokk_code.app import BrokkApp
    except ImportError:
        print("Error: Could not import BrokkApp. Is app.py missing?")
        sys.exit(1)

    session_id = getattr(args, "session", None)
    resume_session = getattr(args, "resume_session", False)

    if args.command == "resume":
        session_id = args.session_id
        resume_session = False  # Explicitly using the provided ID, not "last session" logic

    app = BrokkApp(
        workspace_dir=workspace_path,
        jar_path=jar_path,
        executor_version=args.executor_version,
        executor_snapshot=args.executor_snapshot,
        session_id=session_id,
        resume_session=resume_session,
        vendor=args.vendor,
    )
    app.run()

    # Print resume hint on exit if the session has tasks
    from brokk_code.session_persistence import (
        get_session_zip_path,
        has_tasks,
        load_last_session_id,
    )

    last_id = load_last_session_id(workspace_path)
    if last_id:
        zip_path = get_session_zip_path(workspace_path, last_id)
        if has_tasks(zip_path):
            print(f"brokk resume {last_id}")


if __name__ == "__main__":
    main()



from pathlib import Path

import pytest


@pytest.fixture(autouse=True)
def isolate_home(tmp_path, monkeypatch):
    """Prevent tests from reading/writing real ~/.brokk state and current directory."""
    monkeypatch.setattr(Path, "home", lambda: tmp_path)

    # Isolate CWD to prevent accidental writes to the repo root
    cwd = tmp_path / "cwd"
    cwd.mkdir(parents=True, exist_ok=True)
    monkeypatch.chdir(cwd)



from copy import deepcopy
from typing import Any, Dict, List, Optional

from rich.text import Text
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Vertical, VerticalScroll
from textual.widgets import Label, Static


class TaskListPanel(Vertical):
    """
    Displays the current task list status.

    Note: Currently /v1/context does not expose fragment text content.
    Future enhancement: Add an endpoint to fetch fragment content by ID.
    """

    can_focus = True
    BINDINGS = [
        Binding("left,up", "cursor_prev", "Prev", show=False),
        Binding("right,down", "cursor_next", "Next", show=False),
        Binding("enter,space", "toggle_selected", "Toggle", show=False),
        Binding("a", "task_add", "Add", show=False),
        Binding("e", "task_edit", "Edit", show=False),
        Binding("d", "task_delete", "Delete", show=False),
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._last_details: Optional[Dict[str, Any]] = None
        self._selected_index: int = 0

    @property
    def has_detailed_info(self) -> bool:
        """Returns True if the panel is currently showing detailed data from /v1/tasklist."""
        return self._last_details is not None

    @property
    def selected_index(self) -> int:
        return self._selected_index

    def selected_task(self) -> Optional[Dict[str, Any]]:
        if not self._last_details:
            return None
        tasks = self._last_details.get("tasks", [])
        if not tasks:
            return None
        if self._selected_index < 0 or self._selected_index >= len(tasks):
            return None
        return tasks[self._selected_index]

    def move_selection(self, delta: int) -> bool:
        """Moves selected task row. Returns True when selection changed."""
        if not self._last_details:
            return False
        tasks = self._last_details.get("tasks", [])
        if not tasks:
            return False
        new_index = min(max(0, self._selected_index + delta), len(tasks) - 1)
        if new_index == self._selected_index:
            return False
        self._selected_index = new_index
        self._render_details()
        return True

    def tasklist_data_for_update(self) -> Optional[Dict[str, Any]]:
        """Returns a mutable copy of detailed tasklist data suitable for CRUD updates."""
        if not self._last_details:
            return None
        data = deepcopy(self._last_details)
        data.setdefault("tasks", [])
        return data

    def compose(self) -> ComposeResult:
        yield Label("Task List", id="tasklist-header")
        yield Label("Selected: none", id="tasklist-selection")
        yield Static(self._get_shortcuts_text(), id="tasklist-help-line")
        with VerticalScroll(id="tasklist-container"):
            yield Static("No task list active", id="tasklist-content")

    def _get_shortcuts_text(self) -> str:
        """Derive a concise help line from BINDINGS."""
        shortcuts: list[str] = []
        for binding in self.BINDINGS:
            # Skip navigation and internal-only keys; we add manual entries for these.
            if binding.key in ("left,up", "right,down", "enter,space"):
                continue

            key_display = binding.key.upper()
            if "SHIFT+" in key_display:
                key_display = key_display.replace("SHIFT+", "")
            shortcuts.append(f"[b]{key_display}[/b] {binding.description}")

        manual = [
            "[bold bright_magenta]Esc[/] Close",
            "[b]Up/Down[/b] Move",
            "[b]Space[/b] Toggle",
            "[b]Enter[/b] Toggle",
        ]
        return "  ".join(manual + shortcuts)

    def on_mount(self) -> None:
        self._update_selection_status()

    def action_cursor_prev(self) -> None:
        self.move_selection(-1)

    def action_cursor_next(self) -> None:
        self.move_selection(1)

    def action_toggle_selected(self) -> None:
        app = self.app
        if app is not None and hasattr(app, "action_task_toggle"):
            app.action_task_toggle()

    def action_task_add(self) -> None:
        app = self.app
        if app is not None and hasattr(app, "action_task_add"):
            app.action_task_add()

    def action_task_edit(self) -> None:
        app = self.app
        if app is not None and hasattr(app, "action_task_edit"):
            app.action_task_edit()

    def action_task_delete(self) -> None:
        app = self.app
        if app is not None and hasattr(app, "action_task_delete"):
            app.action_task_delete()

    def refresh_tasklist(self, context_data: Dict[str, Any]) -> None:
        """Finds the TASK_LIST fragment and updates the display using context overview."""
        fragments = context_data.get("fragments", [])
        task_fragment: Optional[Dict[str, Any]] = next(
            (f for f in fragments if f.get("chipKind") == "TASK_LIST"), None
        )

        if not task_fragment:
            self._last_details = None
            self._selected_index = 0
            self._update_selection_status()
            self.query_one("#tasklist-content", Static).update(
                Text("No task list active", style="dim")
            )
            return

        # If we have detailed data already, don't clobber it with the summary
        if self._last_details:
            return

        text = Text()
        text.append("[ ] ", style="bold blue")
        text.append("Task list active", style="bold")
        self.query_one("#tasklist-content", Static).update(text)
        self._update_selection_status()

    def update_tasklist_details(self, tasklist_data: Dict[str, Any]) -> None:
        """Updates the display with detailed task list information from /v1/tasklist."""
        previous_selected_id = ""
        prev = self.selected_task()
        if prev:
            previous_selected_id = str(prev.get("id", ""))
        self._last_details = deepcopy(tasklist_data)
        big_picture = tasklist_data.get("bigPicture")
        tasks = tasklist_data.get("tasks", [])

        content = self.query_one("#tasklist-content", Static)
        if not big_picture and not tasks:
            self._last_details = None
            self._selected_index = 0
            content.update(Text("No task list active", style="dim"))
            self._update_selection_status()
            return

        # Keep the same selected task by id when possible.
        if previous_selected_id:
            for idx, task in enumerate(tasks):
                if str(task.get("id", "")) == previous_selected_id:
                    self._selected_index = idx
                    break
        if tasks:
            self._selected_index = min(max(0, self._selected_index), len(tasks) - 1)
        else:
            self._selected_index = 0

        self._render_details()

    def _update_selection_status(self) -> None:
        selected = self.selected_task()
        label = self.query_one("#tasklist-selection", Label)
        if not selected:
            label.update("Selected: none")
            return
        done = "[x]" if bool(selected.get("done", False)) else "[ ]"
        title = str(selected.get("title", "Task")).strip() or "Task"
        label.update(f"Selected: {done} {title}")

    def _render_details(self) -> None:
        if not self._last_details:
            self.query_one("#tasklist-content", Static).update(
                Text("No task list active", style="dim")
            )
            self._update_selection_status()
            return
        tasks: List[Dict[str, Any]] = self._last_details.get("tasks", [])
        text = Text()
        for i, task in enumerate(tasks, 1):
            done = task.get("done", False)
            title = task.get("title", f"Task {i}")

            checkbox = "[x]" if done else "[ ]"

            prefix = "> " if i - 1 == self._selected_index else "  "
            text.append(prefix, style="bold")
            text.append(f"{checkbox} ", style="bold green" if done else "bold blue")
            text.append(title, style="bold strike" if done else "bold")
            text.append("\n")

        self.query_one("#tasklist-content", Static).update(text)
        self._update_selection_status()



import asyncio
import logging
import random
import re
import time
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional

from textual.app import App, ComposeResult, ScreenStackError
from textual.binding import Binding
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Input, ListItem, ListView, Static

from brokk_code.executor import ExecutorError, ExecutorManager
from brokk_code.prompt_history import append_prompt, clear_history, load_history
from brokk_code.settings import DEFAULT_THEME, Settings, normalize_theme_name
from brokk_code.welcome import build_welcome_message, get_braille_icon
from brokk_code.widgets.chat_panel import ChatInput, ChatPanel
from brokk_code.widgets.context_panel import ContextPanel
from brokk_code.widgets.status_line import StatusLine
from brokk_code.widgets.tasklist_panel import TaskListPanel
from brokk_code.workspace import resolve_workspace_dir

logger = logging.getLogger(__name__)


class ContextModalScreen(ModalScreen[None]):
    """Full-screen modal wrapper for the context panel."""

    BINDINGS = [
        Binding("escape", "close_context", "Close", show=False),
    ]

    def __init__(self, on_close: Callable[[], None]) -> None:
        super().__init__()
        self._on_close = on_close

    def compose(self) -> ComposeResult:
        with Vertical(id="context-modal-container"):
            yield ContextPanel(id="context-panel")

    def on_mount(self) -> None:
        self.query_one(ContextPanel).focus()

    def action_close_context(self) -> None:
        self._on_close()
        self.dismiss(None)


class TaskListModalScreen(ModalScreen[None]):
    """Full-screen modal wrapper for the task list panel."""

    BINDINGS = [
        Binding("escape", "close_tasklist", "Close", show=False),
    ]

    def __init__(self, on_close: Callable[[], None]) -> None:
        super().__init__()
        self._on_close = on_close

    def compose(self) -> ComposeResult:
        with Vertical(id="tasklist-modal-container"):
            yield TaskListPanel(id="tasklist-panel")

    def on_mount(self) -> None:
        self.query_one(TaskListPanel).focus()

    def action_close_tasklist(self) -> None:
        self._on_close()
        self.dismiss(None)


class TaskTitleModalScreen(ModalScreen[Optional[str]]):
    """Small modal to prompt for a task title (add/edit)."""

    BINDINGS = [
        Binding("escape", "dismiss", "Cancel", show=False),
    ]

    def __init__(self, title: str, *, initial: str = "") -> None:
        super().__init__()
        self._title = title
        self._initial = initial

    def compose(self) -> ComposeResult:
        with Vertical(id="task-title-modal-container"):
            yield Static(self._title, id="task-title-modal-title")
            yield Input(value=self._initial, placeholder="Task title", id="task-title-input")

    def on_mount(self) -> None:
        inp = self.query_one("#task-title-input", Input)
        inp.focus()
        inp.cursor_position = len(inp.value)

    def on_input_submitted(self, event: Input.Submitted) -> None:
        value = str(event.value or "").strip()
        self.dismiss(value if value else None)


class ModelSelectModal(ModalScreen[str]):
    """A modal for selecting from available models."""

    BINDINGS = [
        Binding("escape", "dismiss", "Cancel", show=False),
    ]

    def __init__(self, models: List[str]) -> None:
        super().__init__()
        self.models = models
        self._item_id_to_model: Dict[str, str] = {
            f"model-{idx}": model for idx, model in enumerate(models)
        }

    def compose(self) -> ComposeResult:
        with Vertical(id="model-select-container"):
            yield Static("Select Model", id="model-select-title")
            with VerticalScroll(id="model-select-list-wrap"):
                yield ListView(
                    *[
                        ListItem(Static(model_name), id=item_id)
                        for item_id, model_name in self._item_id_to_model.items()
                    ],
                    id="model-select-list",
                )

    def on_mount(self) -> None:
        self.query_one("#model-select-list", ListView).focus()

    def on_list_view_selected(self, message: ListView.Selected) -> None:
        if not message.item or not message.item.id:
            return
        model_name = self._item_id_to_model.get(message.item.id)
        if model_name:
            self.dismiss(model_name)


class ReasoningSelectModal(ModalScreen[str]):
    """A modal for selecting the reasoning level."""

    BINDINGS = [
        Binding("escape", "dismiss", "Cancel", show=False),
    ]

    def __init__(self, levels: List[str], current: str) -> None:
        super().__init__()
        self.levels = levels
        self.current = current
        self._item_id_to_level: Dict[str, str] = {
            f"level-{idx}": level for idx, level in enumerate(levels)
        }

    def compose(self) -> ComposeResult:
        with Vertical(id="reasoning-select-container"):
            yield Static("Select Reasoning", id="reasoning-select-title")
            with VerticalScroll(id="reasoning-select-list-wrap"):
                yield ListView(
                    *[
                        ListItem(
                            Static(
                                f"{'[x]' if level == self.current else '[ ]'} {level}",
                                markup=False,
                            ),
                            id=item_id,
                        )
                        for item_id, level in self._item_id_to_level.items()
                    ],
                    id="reasoning-select-list",
                )

    def on_mount(self) -> None:
        self.query_one("#reasoning-select-list", ListView).focus()

    def on_list_view_selected(self, message: ListView.Selected) -> None:
        if not message.item or not message.item.id:
            return
        level = self._item_id_to_level.get(message.item.id)
        if level:
            self.dismiss(level)


class ModeSelectModal(ModalScreen[str]):
    """A modal for selecting the agent mode."""

    BINDINGS = [
        Binding("escape", "dismiss", "Cancel", show=False),
    ]

    def __init__(self, modes: List[str], current: str) -> None:
        super().__init__()
        self.modes = modes
        self.current = current
        self._item_id_to_mode: Dict[str, str] = {
            f"mode-{idx}": mode for idx, mode in enumerate(modes)
        }

    def compose(self) -> ComposeResult:
        with Vertical(id="mode-select-container"):
            yield Static("Select Mode", id="mode-select-title")
            with VerticalScroll(id="mode-select-list-wrap"):
                yield ListView(
                    *[
                        ListItem(
                            Static(
                                f"{'[x]' if mode == self.current else '[ ]'} {mode}",
                                markup=False,
                            ),
                            id=item_id,
                        )
                        for item_id, mode in self._item_id_to_mode.items()
                    ],
                    id="mode-select-list",
                )

    def on_mount(self) -> None:
        self.query_one("#mode-select-list", ListView).focus()

    def on_list_view_selected(self, message: ListView.Selected) -> None:
        if not message.item or not message.item.id:
            return
        mode = self._item_id_to_mode.get(message.item.id)
        if mode:
            self.dismiss(mode)


class ModelReasoningSelectModal(ModalScreen[tuple[str, str]]):
    """A combined modal for selecting both model and reasoning level side-by-side."""

    BINDINGS = [
        Binding("escape", "dismiss", "Cancel", show=False),
    ]

    def __init__(self, models: List[str], current_model: str, current_reasoning: str) -> None:
        super().__init__()
        self.models = models
        self.selected_model = current_model
        self.selected_reasoning = current_reasoning
        self.reasoning_levels = ["disable", "low", "medium", "high"]

    def compose(self) -> ComposeResult:
        with Horizontal(id="model-reasoning-combined-container"):
            with Vertical(classes="selection-pane"):
                yield Static("Model", id="model-select-title")
                with VerticalScroll(id="model-select-list-wrap"):
                    items = []
                    for idx, m in enumerate(self.models):
                        label = f"{'[x]' if m == self.selected_model else '[ ]'} {m}"
                        items.append(ListItem(Static(label, markup=False), id=f"m-{idx}"))
                    yield ListView(*items, id="model-select-list")

            with Vertical(classes="selection-pane"):
                yield Static("Reasoning", id="reasoning-select-title")
                with VerticalScroll(id="reasoning-select-list-wrap"):
                    items = []
                    for idx, r in enumerate(self.reasoning_levels):
                        label = f"{'[x]' if r == self.selected_reasoning else '[ ]'} {r}"
                        items.append(ListItem(Static(label, markup=False), id=f"r-{idx}"))
                    yield ListView(*items, id="reasoning-select-list")

    def on_mount(self) -> None:
        # Sync model list highlight
        try:
            m_list = self.query_one("#model-select-list", ListView)
            m_idx = self.models.index(self.selected_model)
            m_list.index = m_idx
        except (ValueError, Exception):
            pass

        # Sync reasoning list highlight
        try:
            r_list = self.query_one("#reasoning-select-list", ListView)
            r_idx = self.reasoning_levels.index(self.selected_reasoning)
            r_list.index = r_idx
        except (ValueError, Exception):
            pass

        # Focus the model list by default
        self.query_one("#model-select-list", ListView).focus()

    def on_list_view_selected(self, message: ListView.Selected) -> None:
        if not message.item or not message.item.id:
            return

        try:
            if message.list_view.id == "model-select-list":
                # IDs are 'm-0', 'm-1', etc.
                idx_str = message.item.id.split("-")[-1]
                idx = int(idx_str)
                self.selected_model = self.models[idx]
                # Update markers in model list
                for i, item in enumerate(message.list_view.query(ListItem)):
                    marker = "[x]" if i == idx else "[ ]"
                    # The Static widget was created with markup=False in compose()
                    item.query_one(Static).update(f"{marker} {self.models[i]}")

                # Sync and focus reasoning list
                r_list = self.query_one("#reasoning-select-list", ListView)
                try:
                    r_idx = self.reasoning_levels.index(self.selected_reasoning)
                    r_list.index = r_idx
                except (ValueError, Exception):
                    pass
                r_list.focus()

            elif message.list_view.id == "reasoning-select-list":
                # IDs are 'r-0', 'r-1', etc.
                idx_str = message.item.id.split("-")[-1]
                idx = int(idx_str)
                self.selected_reasoning = self.reasoning_levels[idx]
                # Dismiss immediately upon reasoning selection
                self.dismiss((self.selected_model, self.selected_reasoning))
        except (ValueError, IndexError):
            logger.error("Failed to parse index from ListItem id: %s", message.item.id)


class BrokkApp(App):
    """The main Brokk TUI application.

    Task list UI policy:
    - The task list is accessed via a full-screen modal (TaskListModalScreen).
    - The side task list panel remains mounted for layout stability and potential future use,
      but it is not toggled by /task.
    """

    CSS_PATH = "styles/app.tcss"
    COMMAND_PALETTE_DISPLAY = "Settings"
    BINDINGS = [
        # Footer/help-bar ordering: Context, Tasks, Settings
        Binding("ctrl+c", "handle_ctrl_c", "Quit", show=True),
        Binding("ctrl+p", "command_palette", "Settings", show=True),
        Binding("shift+tab", "toggle_mode", "Toggle mode", show=False, priority=True),
    ]

    def __init__(
        self,
        workspace_dir: Optional[Path] = None,
        jar_path: Optional[Path] = None,
        executor_version: Optional[str] = None,
        executor_snapshot: bool = True,
        executor: Optional[ExecutorManager] = None,
        session_id: Optional[str] = None,
        resume_session: bool = False,
        vendor: Optional[str] = None,
    ) -> None:
        super().__init__()
        if executor:
            self.executor = executor
            if workspace_dir:
                self.executor.workspace_dir = resolve_workspace_dir(workspace_dir)
            if vendor is not None:
                self.executor.vendor = vendor
        else:
            self.executor = ExecutorManager(
                resolve_workspace_dir(workspace_dir or Path.cwd()),
                jar_path,
                executor_version=executor_version,
                executor_snapshot=executor_snapshot,
                vendor=vendor,
            )
        self.requested_session_id = session_id
        self.resume_session = resume_session
        self.settings = Settings.load()
        self._set_theme(self.settings.theme)
        self.agent_mode = "LUTZ"

        # Initialize model and reasoning settings from persisted Settings if present,
        # otherwise fall back to safe defaults.
        # We accept persisted values as-is at startup; validation against the
        # executor model catalog can occur later if needed.
        self.current_model = (
            str(self.settings.last_model).strip()
            if self.settings.last_model and str(self.settings.last_model).strip()
            else "gpt-5.2"
        )
        self.code_model = (
            str(self.settings.last_code_model).strip()
            if self.settings.last_code_model and str(self.settings.last_code_model).strip()
            else "gemini-3-flash-preview"
        )
        self.reasoning_level = (
            str(self.settings.last_reasoning_level).strip()
            if self.settings.last_reasoning_level
            and str(self.settings.last_reasoning_level).strip()
            else "low"
        )
        self.reasoning_level_code = (
            str(self.settings.last_code_reasoning_level).strip()
            if self.settings.last_code_reasoning_level
            else "disable"
        )
        self.auto_commit = (
            bool(self.settings.last_auto_commit)
            if isinstance(self.settings.last_auto_commit, bool)
            else True
        )
        self.current_branch = "unknown"
        self.job_in_progress = False
        self.current_job_id: Optional[str] = None
        self._pending_prompt: Optional[str] = None
        self._pending_updated_at: float = 0
        self._pending_generation: int = 0
        self._pending_min_wait_until: float = 0.0
        self._resubmit_grace_s: float = 0.2
        self._last_ctrl_c_time: float = 0
        self._executor_ready: bool = False
        self._refresh_context_lock = asyncio.Lock()
        self._reported_refresh_errors: set[str] = set()
        self._reasoning_target: str = "planner"

        self._tasklist_restore_focus_widget: Any | None = None

        # Shutdown coordination flags and lock
        self._shutting_down: bool = False
        self._shutdown_completed: bool = False
        self._session_exported: bool = False
        self._shutdown_lock = asyncio.Lock()

    @property
    def current_mode(self) -> str:
        """Alias for agent_mode used by tests and for unified access."""
        return self.agent_mode

    @current_mode.setter
    def current_mode(self, value: str) -> None:
        self.agent_mode = value

    def _maybe_chat(self) -> Optional[ChatPanel]:
        """Safely attempt to get the ChatPanel, returning None if the UI isn't mounted."""
        try:
            return self.query_one(ChatPanel)
        except (ScreenStackError, Exception):
            return None

    def _show_welcome_message(self) -> None:
        """Constructs and displays the branded welcome message in the ChatPanel."""
        chat = self._maybe_chat()
        if not chat:
            return

        chat.add_welcome(get_braille_icon(), build_welcome_message(self.get_slash_commands()))

    def _maybe_statusline(self) -> Optional[StatusLine]:
        """Safely attempt to get the StatusLine, returning None if the UI isn't mounted."""
        try:
            chat = self._maybe_chat()
            if chat:
                return chat.query_one(StatusLine)
            return self.query_one(StatusLine)
        except (ScreenStackError, Exception):
            return None

    def _update_statusline(self) -> None:
        """Collect current state and update the mounted StatusLine (best-effort)."""
        chat = self._maybe_chat()
        if not chat:
            return
        try:
            status = chat.query_one("#status-line", StatusLine)
        except Exception:
            return
        if not status:
            return
        try:
            workspace = None
            try:
                if getattr(self, "executor", None) is not None:
                    ws = getattr(self.executor, "workspace_dir", None)
                    if ws is not None:
                        workspace = str(ws)
            except Exception:
                workspace = None

            status.update_status(
                mode=getattr(self, "current_mode", getattr(self, "agent_mode", "unknown")),
                model=getattr(self, "current_model", None),
                reasoning=getattr(self, "reasoning_level", None),
                workspace=workspace,
                branch=getattr(self, "current_branch", "unknown"),
            )
        except Exception:
            # Swallow all errors when updating UI that's possibly not mounted in tests.
            return

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield ChatPanel(id="chat-main")
            yield TaskListPanel(id="side-tasklist")

    async def on_mount(self) -> None:
        chat = self._maybe_chat()
        logger.info("Using workspace directory: %s", self.executor.workspace_dir)
        if chat:
            chat.set_token_bar_visible(True)

            # Load initial prompt history for arrow-key navigation
            history = load_history(self.executor.workspace_dir)
            chat.set_history(history)

            self._show_welcome_message()
            chat.add_system_message("Starting Brokk executor...")

        self.run_worker(self._start_executor())
        self.run_worker(self._monitor_executor())
        self.run_worker(self._poll_tasklist())
        self.run_worker(self._poll_context())
        self._update_statusline()

    async def _start_executor(self) -> None:
        chat = self._maybe_chat()
        try:
            from brokk_code.session_persistence import (
                get_session_zip_path,
                load_last_session_id,
                save_last_session_id,
            )

            await self.executor.start()

            # Fetch and display effective build hint immediately
            try:
                live_info = await self.executor.get_health_live()
                version = live_info.get("version", "unknown")
                proto = live_info.get("protocolVersion", "unknown")
                eid = live_info.get("execId", "unknown")
                msg = f"Connected to executor {eid} (version: {version}, protocol: {proto})"
                if chat:
                    chat.add_system_message(msg)
                else:
                    logger.info(msg)
            except Exception:
                logger.debug("Failed to fetch health/live info", exc_info=True)

            # Session Management Logic
            session_to_resume = self.requested_session_id
            if not session_to_resume and self.resume_session:
                session_to_resume = load_last_session_id(self.executor.workspace_dir)

            resumed = False
            if session_to_resume:
                zip_path = get_session_zip_path(self.executor.workspace_dir, session_to_resume)
                if zip_path.exists():
                    try:
                        msg = f"Resuming session {session_to_resume}..."
                        if chat:
                            chat.add_system_message(msg)
                        else:
                            logger.info(msg)
                        zip_bytes = zip_path.read_bytes()
                        await self.executor.import_session_zip(
                            zip_bytes, session_id=session_to_resume
                        )
                        resumed = True
                    except Exception as e:
                        logger.warning("Failed to resume session %s: %s", session_to_resume, e)

            if not resumed:
                await self.executor.create_session()

            if self.executor.session_id:
                save_last_session_id(self.executor.workspace_dir, self.executor.session_id)

            if await self.executor.wait_ready():
                self._executor_ready = True
                # Initial context load
                self.run_worker(self._refresh_context_panel())
            else:
                msg = "Executor failed to become ready (timeout)."
                if chat:
                    chat.add_system_message(msg, level="ERROR")
                else:
                    logger.error(msg)
        except ExecutorError as e:
            msg = str(e)
            if "jbang" in msg.lower():
                msg += (
                    "\n\nHint: Install jbang from https://jbang.dev "
                    "or provide a local JAR with --jar."
                )
            if chat:
                chat.add_system_message(msg, level="ERROR")
            else:
                logger.error(msg)
        except Exception as e:
            msg = f"Unexpected startup error: {e}"
            if chat:
                chat.add_system_message(msg, level="ERROR")
            else:
                logger.error(msg)

    async def _monitor_executor(self) -> None:
        """Background worker to check if the executor dies unexpectedly."""
        while True:
            await asyncio.sleep(2.0)
            if not self.executor.check_alive():
                msg = "Executor process crashed unexpectedly."
                chat = self._maybe_chat()
                if chat:
                    chat.add_system_message(msg, level="ERROR")
                else:
                    logger.error(msg)
                break

    async def _poll_tasklist(self) -> None:
        """Periodically refreshes the task list details."""
        while True:
            if self._executor_ready:
                # We poll even if a job is running, as /v1/tasklist is low impact
                try:
                    tasklist_data = await self.executor.get_tasklist()
                    self._update_tasklist_details_all(tasklist_data)
                except Exception:
                    logger.debug("Periodic tasklist poll failed", exc_info=True)
            await asyncio.sleep(15.0)

    async def _poll_context(self) -> None:
        """Periodically refreshes the context panel."""
        while True:
            if self._executor_ready:
                # refresh_context_panel handles both ContextPanel and TaskListPanel overview
                await self._refresh_context_panel()
            # Sleep 10-15s with jitter
            await asyncio.sleep(random.uniform(10.0, 15.0))

    async def _refresh_context_panel(self) -> None:
        """Fetches latest context and updates context, task list, and chat panels."""
        if not self._executor_ready:
            return

        async with self._refresh_context_lock:
            try:
                context_data = await self.executor.get_context()
                self.current_branch = context_data.get("branch", "unknown")

                # UI updates are best-effort if screen is not on stack
                try:
                    if isinstance(self.screen, ContextModalScreen):
                        self.screen.query_one(ContextPanel).refresh_context(context_data)
                    else:
                        self.query_one(ContextPanel).refresh_context(context_data)
                except (ScreenStackError, Exception):
                    pass

                try:
                    for task_list in self._tasklist_panels():
                        if not task_list.has_detailed_info:
                            task_list.refresh_tasklist(context_data)
                except (ScreenStackError, Exception):
                    pass

                # Update token usage in ChatPanel
                chat = self._maybe_chat()
                if chat:
                    used = context_data.get("usedTokens", 0)
                    max_tokens = context_data.get("maxTokens")
                    fragments = context_data.get("fragments")
                    chat.set_token_usage(used, max_tokens, fragments)

                self._update_statusline()

                # Clear error tracking on success
                self._reported_refresh_errors.clear()
            except Exception as e:
                # Rate-limit notifications to once per unique exception type per session
                err_key = type(e).__name__
                if err_key not in self._reported_refresh_errors:
                    msg = f"Context refresh failed: {e}"
                    chat = self._maybe_chat()
                    if chat:
                        chat.add_system_message(msg, level="ERROR")
                    else:
                        logger.error(msg)
                    self._reported_refresh_errors.add(err_key)
                logger.debug("Failed to refresh context panel", exc_info=True)

    def on_context_panel_action_requested(self, message: ContextPanel.ActionRequested) -> None:
        self.run_worker(self._execute_context_action(message))

    async def _execute_context_action(self, message: ContextPanel.ActionRequested) -> None:
        if not self._executor_ready:
            return

        chat = self._maybe_chat()
        if isinstance(self.screen, ContextModalScreen):
            panel = self.screen.query_one(ContextPanel)
        else:
            panel = self.query_one(ContextPanel)
        selected_fragments = panel.selected_fragments
        try:
            match message.action:
                case "drop_selected":
                    await self.executor.drop_context_fragments(message.fragment_ids)
                    if chat:
                        chat.add_system_message(f"Dropped {len(message.fragment_ids)} fragment(s).")
                case "drop_others":
                    to_drop = self._compute_drop_others(
                        panel._fragments, selected_fragments, panel._active_id
                    )
                    if to_drop:
                        await self.executor.drop_context_fragments(to_drop)
                        if chat:
                            chat.add_system_message(f"Dropped {len(to_drop)} other fragment(s).")
                case "drop_all":
                    await self.executor.drop_all_context()
                    if chat:
                        chat.add_system_message("Dropped all context fragments.")
                case "toggle_pin_selected":
                    updates = self._collect_pin_updates(selected_fragments)
                    for fragment_id, pinned in updates:
                        await self.executor.set_context_fragment_pinned(fragment_id, pinned)
                    if chat and updates:
                        chat.add_system_message(
                            f"Updated pin state for {len(updates)} fragment(s)."
                        )
                case "toggle_readonly_selected":
                    updates = self._collect_readonly_updates(selected_fragments)
                    for fragment_id, readonly in updates:
                        await self.executor.set_context_fragment_readonly(fragment_id, readonly)
                    if chat and updates:
                        chat.add_system_message(
                            f"Updated read-only state for {len(updates)} editable fragment(s)."
                        )
                case "compress_history":
                    await self.executor.compress_context_history()
                    if chat:
                        chat.add_system_message("Compressing history...")
                case "clear_history":
                    await self.executor.clear_context_history()
                    if chat:
                        chat.add_system_message("Cleared history.")
                case _:
                    logger.warning("Unknown context action requested: %s", message.action)
                    return

            await self._refresh_context_panel()
        except Exception as e:
            if chat:
                chat.add_system_message(f"Context action failed: {e}", level="ERROR")
            else:
                logger.error("Context action failed: %s", e)

    @staticmethod
    def _compute_drop_others(
        all_fragments: List[Dict[str, Any]],
        selected_fragments: List[Dict[str, Any]],
        active_id: Optional[str],
    ) -> List[str]:
        protected_ids = {str(f.get("id", "")) for f in selected_fragments}
        if active_id:
            protected_ids.add(active_id)

        to_drop = []
        for f in all_fragments:
            f_id = str(f.get("id", ""))
            if not f_id or f_id in protected_ids:
                continue
            if bool(f.get("pinned", False)):
                continue
            chip_kind = str(f.get("chip_kind", f.get("chipKind", ""))).upper()
            if chip_kind == "HISTORY":
                continue
            to_drop.append(f_id)
        return to_drop

    @staticmethod
    def _collect_pin_updates(selected_fragments: List[Dict[str, Any]]) -> List[tuple[str, bool]]:
        updates: List[tuple[str, bool]] = []
        for fragment in selected_fragments:
            fragment_id = str(fragment.get("id", "")).strip()
            if not fragment_id:
                continue
            current = bool(fragment.get("pinned", False))
            updates.append((fragment_id, not current))
        return updates

    @staticmethod
    def _collect_readonly_updates(
        selected_fragments: List[Dict[str, Any]],
    ) -> List[tuple[str, bool]]:
        updates: List[tuple[str, bool]] = []
        for fragment in selected_fragments:
            if not bool(fragment.get("editable", False)):
                continue
            fragment_id = str(fragment.get("id", "")).strip()
            if not fragment_id:
                continue
            current = bool(fragment.get("readonly", False))
            updates.append((fragment_id, not current))
        return updates

    def _tasklist_panels(self) -> List[TaskListPanel]:
        """Return all mounted task list panels (side + modal if present)."""
        panels: List[TaskListPanel] = []
        try:
            panels.append(self.query_one("#side-tasklist", TaskListPanel))
        except Exception:
            pass

        try:
            if isinstance(self.screen, TaskListModalScreen):
                panels.append(self.screen.query_one(TaskListPanel))
        except Exception:
            pass

        return panels

    def _active_tasklist_panel(self) -> TaskListPanel:
        """Return the panel that should receive task actions."""
        try:
            current_screen = self.screen
        except ScreenStackError:
            current_screen = None

        if isinstance(current_screen, TaskListModalScreen):
            try:
                return current_screen.query_one(TaskListPanel)
            except Exception:
                # Modal may be current but its contents not mounted yet; fall back to side panel.
                return self.query_one("#side-tasklist", TaskListPanel)

        return self.query_one("#side-tasklist", TaskListPanel)

    def _update_tasklist_details_all(self, tasklist_data: Dict[str, Any]) -> None:
        for panel in self._tasklist_panels():
            panel.update_tasklist_details(tasklist_data)

    async def _ensure_tasklist_data(self) -> Optional[Dict[str, Any]]:
        panel = self._active_tasklist_panel()
        data = panel.tasklist_data_for_update()
        if data is not None:
            return data
        data = await self.executor.get_tasklist()
        self._update_tasklist_details_all(data)
        return panel.tasklist_data_for_update()

    async def _persist_tasklist(self, data: Dict[str, Any]) -> Dict[str, Any]:
        saved = await self.executor.set_tasklist(data)
        self._update_tasklist_details_all(saved)
        return saved

    async def _toggle_selected_task(self) -> None:
        chat = self.query_one(ChatPanel)
        panel = self._active_tasklist_panel()
        selected = panel.selected_task()
        if not selected:
            chat.add_system_message("No task selected.")
            return

        task_id = str(selected.get("id", "")).strip()
        if not task_id:
            chat.add_system_message("Selected task has no ID and cannot be updated.", level="ERROR")
            return

        try:
            data = await self._ensure_tasklist_data()
            if not data:
                chat.add_system_message("No task list active.")
                return
            tasks = data.get("tasks", [])
            for task in tasks:
                if str(task.get("id", "")).strip() == task_id:
                    task["done"] = not bool(task.get("done", False))
                    break
            await self._persist_tasklist(data)
        except Exception as e:
            chat.add_system_message(f"Failed to toggle task: {e}", level="ERROR")

    async def _delete_selected_task(self) -> None:
        chat = self.query_one(ChatPanel)
        panel = self._active_tasklist_panel()
        selected = panel.selected_task()
        if not selected:
            chat.add_system_message("No task selected.")
            return

        task_id = str(selected.get("id", "")).strip()
        if not task_id:
            chat.add_system_message("Selected task has no ID and cannot be deleted.", level="ERROR")
            return

        try:
            data = await self._ensure_tasklist_data()
            if not data:
                chat.add_system_message("No task list active.")
                return
            before = len(data.get("tasks", []))
            data["tasks"] = [
                task for task in data.get("tasks", []) if str(task.get("id", "")).strip() != task_id
            ]
            if len(data["tasks"]) == before:
                chat.add_system_message("Selected task no longer exists.")
                return
            await self._persist_tasklist(data)
        except Exception as e:
            chat.add_system_message(f"Failed to delete task: {e}", level="ERROR")

    async def _add_task(self, title: str) -> None:
        chat = self.query_one(ChatPanel)
        normalized_title = title.strip()
        if not normalized_title:
            chat.add_system_message("Task title cannot be blank.", level="ERROR")
            return
        try:
            data = await self._ensure_tasklist_data()
            if not data:
                data = {"bigPicture": None, "tasks": []}
            tasks = data.get("tasks", [])
            tasks.append({"title": normalized_title, "text": normalized_title, "done": False})
            data["tasks"] = tasks
            await self._persist_tasklist(data)
        except Exception as e:
            chat.add_system_message(f"Failed to add task: {e}", level="ERROR")

    async def _edit_selected_task(self, title: str) -> None:
        chat = self.query_one(ChatPanel)
        panel = self._active_tasklist_panel()
        selected = panel.selected_task()
        if not selected:
            chat.add_system_message("No task selected.")
            return

        task_id = str(selected.get("id", "")).strip()
        if not task_id:
            chat.add_system_message("Selected task has no ID and cannot be updated.", level="ERROR")
            return

        normalized_title = title.strip()
        if not normalized_title:
            chat.add_system_message("Task title cannot be blank.", level="ERROR")
            return

        try:
            data = await self._ensure_tasklist_data()
            if not data:
                chat.add_system_message("No task list active.")
                return
            updated = False
            for task in data.get("tasks", []):
                if str(task.get("id", "")).strip() == task_id:
                    task["title"] = normalized_title
                    if not str(task.get("text", "")).strip():
                        task["text"] = normalized_title
                    updated = True
                    break
            if not updated:
                chat.add_system_message("Selected task no longer exists.")
                return
            await self._persist_tasklist(data)
        except Exception as e:
            chat.add_system_message(f"Failed to edit task: {e}", level="ERROR")

    def on_chat_panel_mode_selected(self, message: ChatPanel.ModeSelected) -> None:
        self._set_mode(message.mode.upper())

    def on_chat_panel_reasoning_level_selected(
        self, message: ChatPanel.ReasoningLevelSelected
    ) -> None:
        chat = self._maybe_chat()
        if self._reasoning_target == "code":
            self.reasoning_level_code = message.level
            try:
                self.settings.last_code_reasoning_level = message.level
                self.settings.save()
            except Exception:
                logger.exception("Failed to persist code reasoning level")
            if chat:
                chat.add_system_message_markup(
                    f"Code reasoning level changed to: [bold]{message.level}[/]"
                )
        else:
            self.reasoning_level = message.level
            try:
                self.settings.last_reasoning_level = message.level
                self.settings.save()
            except Exception:
                logger.exception("Failed to persist reasoning level")
            if chat:
                chat.add_system_message_markup(
                    f"Reasoning level changed to: [bold]{message.level}[/]"
                )

        self._update_statusline()

    def on_chat_panel_submitted(self, message: ChatPanel.Submitted) -> None:
        """
        Handles user input from the chat panel.

        Persistence Policy:
        - Only non-command prompts (text not starting with '/') are recorded.
        - Prompts are recorded at the moment of submission, regardless of whether
        they trigger a cancellation or are later aborted.
        - History is stored in the project-specific directory:
          `self.executor.workspace_dir / ".brokk" / "prompts.json"`
        """
        raw_text = message.text
        check_text = raw_text.strip()

        if check_text.startswith("/"):
            self._handle_command(check_text)
        elif check_text:
            # Record in history regardless of routing
            append_prompt(
                self.executor.workspace_dir, raw_text, max_history=self.settings.prompt_history_size
            )
            chat = self.query_one(ChatPanel)
            chat.add_history_entry(raw_text)

            chat.add_user_message(raw_text)
            if self.job_in_progress and self.current_job_id:
                self._pending_prompt = raw_text
                now = time.monotonic()
                self._pending_updated_at = now
                self._pending_generation += 1
                self._pending_min_wait_until = max(
                    self._pending_min_wait_until, now + self._resubmit_grace_s
                )
                # Avoid redundant cancellation messages if already pending
                if self._pending_generation == 1:
                    chat.add_system_message("Interrupting current job to start new request...")
                self.run_worker(self.executor.cancel_job(self.current_job_id))
            else:
                self.run_worker(self._run_job(raw_text))

    @staticmethod
    def _extract_at_mentions(task_input: str) -> List[str]:
        """Extracts whitespace-delimited @mention tokens from prompt text."""
        tokens = re.findall(r"(?<!\S)@([^\s@]+)", task_input)
        unique_tokens: List[str] = []
        seen = set()
        for token in tokens:
            norm = token.strip()
            if not norm or norm in seen:
                continue
            seen.add(norm)
            unique_tokens.append(norm)
        return unique_tokens

    @staticmethod
    def _extract_fragment_ids_from_add_context_response(resp: Any) -> List[str]:
        if not isinstance(resp, dict):
            return []

        added = resp.get("added")
        if not isinstance(added, list):
            return []

        ids: List[str] = []
        for item in added:
            if not isinstance(item, dict):
                continue
            raw_id = item.get("id", item.get("fragmentId"))
            if raw_id is None:
                continue
            frag_id = str(raw_id).strip()
            if frag_id:
                ids.append(frag_id)

        return list(dict.fromkeys(ids))

    async def _attach_mentions_to_context(self, task_input: str) -> List[str]:
        """Resolves @mentions and attaches matching entities to context before job submission.

        Returns fragment IDs for any newly-attached context fragments when the executor's
        add_context_* endpoints include them in their payload. If the executor does not
        return fragment IDs, rollback-on-submit-failure is not possible.
        """
        mentions = self._extract_at_mentions(task_input)
        if not mentions:
            return []
        if not hasattr(self.executor, "get_completions"):
            return []

        file_paths: List[str] = []
        class_names: List[str] = []
        method_names: List[str] = []

        for mention in mentions:
            try:
                completion_data = await self.executor.get_completions(mention, limit=20)
            except Exception:
                logger.exception("Failed resolving @mention '%s' via completions", mention)
                continue

            raw_items = completion_data.get("completions", [])
            if not isinstance(raw_items, list):
                continue

            selected: Optional[Dict[str, str]] = None
            mention_lower = mention.lower()
            for raw in raw_items:
                if not isinstance(raw, dict):
                    continue
                detail = str(raw.get("detail", "")).strip()
                name = str(raw.get("name", "")).strip()

                if detail.lower() == mention_lower or name.lower() == mention_lower:
                    selected = {
                        "type": str(raw.get("type", "")).strip().lower(),
                        "detail": detail,
                        "name": name,
                    }
                    break

            if selected is None:
                chat = self._maybe_chat()
                if chat:
                    chat.add_system_message(f"No exact match for @{mention}")
                continue

            completion_type = selected["type"]
            detail = selected["detail"] or selected["name"]
            if not detail:
                continue

            if completion_type == "file":
                file_paths.append(detail)
            elif completion_type in {"class", "module"}:
                class_names.append(detail)
            elif completion_type == "function":
                method_names.append(detail)
            elif completion_type == "field":
                if "." in detail:
                    class_names.append(detail.rsplit(".", 1)[0])

        # De-duplicate while preserving order
        file_paths = list(dict.fromkeys(file_paths))
        class_names = list(dict.fromkeys(class_names))
        method_names = list(dict.fromkeys(method_names))

        attached_parts: List[str] = []
        attached_fragment_ids: List[str] = []

        if file_paths and hasattr(self.executor, "add_context_files"):
            try:
                resp = await self.executor.add_context_files(file_paths)
                attached_fragment_ids.extend(
                    self._extract_fragment_ids_from_add_context_response(resp)
                )
                attached_parts.append(f"files={len(file_paths)}")
            except Exception:
                logger.exception("Failed attaching @mentions as context files")
        if class_names and hasattr(self.executor, "add_context_classes"):
            try:
                resp = await self.executor.add_context_classes(class_names)
                attached_fragment_ids.extend(
                    self._extract_fragment_ids_from_add_context_response(resp)
                )
                attached_parts.append(f"classes={len(class_names)}")
            except Exception:
                logger.exception("Failed attaching @mentions as context classes")
        if method_names and hasattr(self.executor, "add_context_methods"):
            try:
                resp = await self.executor.add_context_methods(method_names)
                attached_fragment_ids.extend(
                    self._extract_fragment_ids_from_add_context_response(resp)
                )
                attached_parts.append(f"methods={len(method_names)}")
            except Exception:
                logger.exception("Failed attaching @mentions as context methods")

        if attached_parts:
            chat = self._maybe_chat()
            if chat:
                details = ", ".join(attached_parts)
                chat.add_system_message(f"Attached @mentions to context: {details}")

        return list(dict.fromkeys(attached_fragment_ids))

    async def _run_job(self, task_input: str) -> None:
        self.job_in_progress = True
        chat = self.query_one(ChatPanel)
        chat.set_job_running(True)
        chat.set_response_pending()
        attached_fragment_ids: List[str] = []
        try:
            attached_fragment_ids = await self._attach_mentions_to_context(task_input)
            self.current_job_id = await self.executor.submit_job(
                task_input,
                self.current_model,
                code_model=self.code_model,
                reasoning_level=self.reasoning_level,
                reasoning_level_code=self.reasoning_level_code,
                mode=self.current_mode,
                auto_commit=self.auto_commit,
            )
            async for event in self.executor.stream_events(self.current_job_id):
                self._handle_event(event)
        except Exception as e:
            if (
                self.current_job_id is None
                and attached_fragment_ids
                and hasattr(self.executor, "drop_context_fragments")
            ):
                try:
                    await self.executor.drop_context_fragments(attached_fragment_ids)
                except Exception:
                    logger.exception(
                        "Failed to rollback context fragments after submit_job failure: %s",
                        attached_fragment_ids,
                    )

            chat.add_system_message(f"Job failed or network error: {e}", level="ERROR")
        finally:
            chat.set_response_finished()
            chat.set_job_running(False)

            # Yield to the event loop to allow any rapid subsequent submissions
            # triggered by the cancellation to be processed before we check _pending_prompt.
            await asyncio.sleep(0)

            if self._pending_prompt:
                # Wait for both the grace window (since cancellation)
                # and the stability debounce (since last keystroke/submit).
                debounce_window = 0.05  # 50ms
                while True:
                    now = time.monotonic()
                    current_gen = self._pending_generation
                    elapsed_since_update = now - self._pending_updated_at

                    # We must be past the absolute grace timestamp AND stable
                    # for the debounce window
                    if (
                        now >= self._pending_min_wait_until
                        and elapsed_since_update >= debounce_window
                        and self._pending_generation == current_gen
                    ):
                        break
                    await asyncio.sleep(0.01)

                next_prompt = self._pending_prompt
                self._pending_prompt = None
                self._pending_updated_at = 0
                self._pending_generation = 0
                self._pending_min_wait_until = 0.0

                # Recurse within the same worker context to prevent
                # the app from flickering to 'idle' and allowing race-condition submits.
                # We keep job_in_progress = True during this transition.
                if next_prompt:
                    await self._run_job(next_prompt)
                else:
                    self.job_in_progress = False
                    self.current_job_id = None
            else:
                # Only mark idle once we are sure no more prompts are queued
                self.job_in_progress = False
                self.current_job_id = None

    def _handle_event(self, event: Dict[str, Any]) -> None:
        event_type = event.get("type")
        data = event.get("data", {})
        chat = self.query_one(ChatPanel)

        if event_type == "LLM_TOKEN":
            chat.append_token(
                token=data.get("token", ""),
                message_type=data.get("messageType", "AI"),
                is_new_message=bool(data.get("isNewMessage", False)),
                is_reasoning=bool(data.get("isReasoning", False)),
                is_terminal=bool(data.get("isTerminal", False)),
            )
        elif event_type == "NOTIFICATION":
            level = data.get("level", "INFO")
            msg = data.get("message", "")
            chat.add_system_message(msg, level=level)
        elif event_type == "ERROR":
            msg = data.get("message", "Unknown error")
            chat.add_system_message(msg, level="ERROR")
            # Note: set_job_running(False) happens in _run_job finally block
        elif event_type == "STATE_HINT":
            hint_name = data.get("name")
            if hint_name in ("contextHistoryUpdated", "workspaceUpdated"):
                self.run_worker(self._refresh_context_panel())

    def _set_mode(self, new_mode: str, *, announce: bool = True) -> None:
        """Sets the agent mode, updates the status line, and optionally announces to chat."""
        self.agent_mode = new_mode
        # Update statusline if present
        self._update_statusline()
        if announce:
            msg_markup = f"Mode changed to: [bold]{self.agent_mode}[/]"
            chat = self._maybe_chat()
            if chat:
                chat.add_system_message_markup(msg_markup)
            else:
                logger.info("Mode changed to %s", self.agent_mode)

    def _render_info(self) -> None:
        """Renders current status and configuration info to the chat."""
        chat = self.query_one(ChatPanel)
        status = (
            "[bold green]Ready[/]" if self._executor_ready else "[bold yellow]Initializing...[/]"
        )
        jar_path = self.executor.resolved_jar_path or "via jbang"
        launch_mode = "Direct JAR" if self.executor.resolved_jar_path else "jbang"

        planner_info = (
            f"Planner Model: [bold]{self.current_model}[/] "
            f"(reasoning: [bold]{self.reasoning_level}[/])"
        )
        code_info = (
            f"Code Model: [bold]{self.code_model}[/] "
            f"(reasoning: [bold]{self.reasoning_level_code}[/])"
        )
        info_markup = (
            f"Status: {status}\n"
            f"Workspace: [bold]{self.executor.workspace_dir}[/]\n"
            f"Launch Mode: [bold]{launch_mode}[/]\n"
            f"Executor JAR: [bold]{jar_path}[/]\n"
            f"Mode: [bold]{self.agent_mode}[/]\n"
            f"Auto-commit: [bold]{'ON' if self.auto_commit else 'OFF'}[/]\n"
            f"{planner_info}\n"
            f"{code_info}"
        )
        chat.add_system_message_markup(info_markup)

    @staticmethod
    def get_slash_commands() -> List[Dict[str, str]]:
        """Returns the structured catalog of supported slash commands."""
        return [
            {"command": "/context", "description": "Toggle and focus context panel"},
            {"command": "/code", "description": "Set mode to CODE (direct implementation)"},
            {"command": "/ask", "description": "Set mode to ASK (questions only)"},
            {"command": "/lutz", "description": "Set mode to LUTZ (default; full agent access)"},
            {"command": "/mode", "description": "Open mode selection menu"},
            {"command": "/model", "description": "Change the planner LLM model"},
            {"command": "/model-code", "description": "Change the code LLM model"},
            {"command": "/reasoning", "description": "Set reasoning level for planner"},
            {"command": "/reasoning-code", "description": "Set reasoning level for code model"},
            {"command": "/autocommit", "description": "Toggle auto-commit for submitted jobs"},
            {"command": "/settings", "description": "Open settings"},
            {"command": "/history", "description": "Show recent prompt history"},
            {"command": "/history-clear", "description": "Clear prompt history"},
            {"command": "/task", "description": "Open/close the task list"},
            {"command": "/info", "description": "Show current configuration and status"},
            {"command": "/help", "description": "Show help message"},
            {"command": "/quit", "description": "Exit the application"},
            {"command": "/exit", "description": "Exit the application"},
        ]

    def _handle_command(self, cmd: str) -> None:
        chat = self.query_one(ChatPanel)
        parts = cmd.split()
        base = parts[0].lower()

        if base == "/model":
            if len(parts) > 1:
                self.current_model = parts[1]
                # Persist the last-used planner model for subsequent runs
                try:
                    self.settings.last_model = self.current_model
                    self.settings.save()
                except Exception:
                    logger.exception("Failed to persist last_model setting")
                chat.add_system_message_markup(f"Model changed to: [bold]{self.current_model}[/]")
                self._update_statusline()
            else:
                self.run_worker(self.action_select_model_and_reasoning())
        elif base == "/model-code":
            if len(parts) > 1:
                self.code_model = parts[1]
                # Persist the last-used code model
                try:
                    self.settings.last_code_model = self.code_model
                    self.settings.save()
                except Exception:
                    logger.exception("Failed to persist last_code_model setting")
                chat.add_system_message_markup(f"Code model changed to: [bold]{self.code_model}[/]")
                self._update_statusline()
            else:
                self.run_worker(self.action_select_code_model_and_reasoning())
        elif base == "/reasoning":
            if len(parts) > 1:
                self.reasoning_level = parts[1]
                # Persist planner reasoning preference
                try:
                    self.settings.last_reasoning_level = self.reasoning_level
                    self.settings.save()
                except Exception:
                    logger.exception("Failed to persist last_reasoning_level setting")
                chat.add_system_message_markup(
                    f"Reasoning level changed to: [bold]{self.reasoning_level}[/]"
                )
                self._update_statusline()
            else:
                self.action_select_reasoning(target="planner")
        elif base == "/reasoning-code":
            if len(parts) > 1:
                self.reasoning_level_code = parts[1]
                # Persist code reasoning preference
                try:
                    self.settings.last_code_reasoning_level = self.reasoning_level_code
                    self.settings.save()
                except Exception:
                    logger.exception("Failed to persist last_code_reasoning_level setting")
                chat.add_system_message_markup(
                    f"Code reasoning level changed to: [bold]{self.reasoning_level_code}[/]"
                )
                self._update_statusline()
            else:
                self.action_select_reasoning(target="code")
        elif base == "/autocommit":
            if len(parts) == 1:
                state = "ON" if self.auto_commit else "OFF"
                chat.add_system_message_markup(
                    f"Auto-commit mode: [bold]{state}[/]\nUsage: /autocommit on|off|toggle",
                    level="WARNING",
                )
                return

            if len(parts) != 2:
                chat.add_system_message(
                    "Usage: /autocommit on|off|toggle (or true/false/1/0/yes/no)",
                    level="ERROR",
                )
                return

            arg = parts[1].strip().lower()
            truthy = {"on", "true", "1", "yes"}
            falsy = {"off", "false", "0", "no"}
            if arg in truthy:
                new_value = True
            elif arg in falsy:
                new_value = False
            elif arg == "toggle":
                new_value = not self.auto_commit
            else:
                chat.add_system_message(
                    "Usage: /autocommit on|off|toggle (or true/false/1/0/yes/no)",
                    level="ERROR",
                )
                return

            self.auto_commit = new_value
            try:
                self.settings.last_auto_commit = self.auto_commit
                self.settings.save()
            except Exception:
                logger.exception("Failed to persist last_auto_commit setting")

            if self.auto_commit:
                chat.add_system_message_markup("Auto-commit mode: [bold]ON[/]")
            else:
                chat.add_system_message_markup(
                    "Auto-commit mode: [bold]OFF[/] (changes will not be committed automatically)",
                    level="WARNING",
                )
        elif base == "/settings":
            if len(parts) > 1:
                chat.add_system_message("Settings opens from /settings with no arguments.")
            self.action_command_palette()
        elif base in ("/code", "/ask", "/lutz"):
            self._set_mode(base[1:].upper())
        elif base == "/mode":
            if len(parts) > 1:
                self._set_mode(parts[1].upper())
            else:
                self.action_select_mode()
        elif base == "/info":
            self._render_info()
        elif base == "/history":
            history = load_history(self.executor.workspace_dir)
            if not history:
                chat.add_system_message("Prompt history is empty.")
            else:
                formatted = "\n".join(f"{i + 1}. {p}" for i, p in enumerate(history))
                chat.append_message("System", f"Recent Prompts:\n{formatted}")
        elif base == "/history-clear":
            clear_history(self.executor.workspace_dir)
            chat.set_history([])
            chat.add_system_message("Prompt history cleared.")
        elif base == "/context":
            self.action_toggle_context()
        elif base == "/task":
            if len(parts) != 1:
                chat.add_system_message(
                    "Usage: /task (task actions are available via task list keybindings).",
                    level="WARNING",
                )
                return
            self.action_toggle_tasklist()
        elif base == "/help":
            commands = self.get_slash_commands()
            # Calculate padding based on longest command
            max_cmd_len = max(len(c["command"]) for c in commands)
            lines = ["Available commands:"]
            for c in commands:
                lines.append(f"  {c['command']: <{max_cmd_len}} - {c['description']}")
            chat.append_message("System", "\n".join(lines))
        elif base in ("/quit", "/exit"):
            self.action_quit()
        else:
            chat.append_message("System", f"Unknown command: {base}. Type /help for assistance.")

    async def action_select_model(self) -> None:
        chat = self._maybe_chat()
        if not self._executor_ready:
            if chat:
                chat.add_system_message(
                    "Executor is not ready. Cannot select model.", level="ERROR"
                )
            return

        try:
            models_data = await self.executor.get_models()
            raw_models = models_data.get("models", [])
            if not isinstance(raw_models, list):
                raw_models = []
            available_models: List[str] = []
            for model in raw_models:
                if isinstance(model, str):
                    name = model.strip()
                elif isinstance(model, dict):
                    name = str(model.get("name", "")).strip()
                else:
                    name = ""
                if name:
                    available_models.append(name)
            if not available_models:
                if chat:
                    chat.add_system_message("No models available from executor.", level="ERROR")
                return

            def update_selection(model_id: str | None) -> None:
                if model_id:
                    self.current_model = model_id
                    # Persist choice
                    try:
                        self.settings.last_model = model_id
                        self.settings.save()
                    except Exception:
                        logger.exception("Failed to persist model setting")

                    if chat:
                        chat.add_system_message_markup(f"Model changed to: [bold]{model_id}[/]")

                    # Update statusline (best-effort)
                    try:
                        self._update_statusline()
                    except Exception:
                        pass

            self.push_screen(ModelSelectModal(available_models), update_selection)
        except Exception as e:
            if chat:
                chat.add_system_message(f"Failed to fetch models: {e}", level="ERROR")

    async def action_select_model_and_reasoning(self) -> None:
        await self._select_model_and_reasoning_flow(target="planner")

    async def action_select_code_model_and_reasoning(self) -> None:
        await self._select_model_and_reasoning_flow(target="code")

    async def _select_model_and_reasoning_flow(self, target: str = "planner") -> None:
        chat = self._maybe_chat()
        if not self._executor_ready:
            if chat:
                chat.add_system_message(
                    "Executor is not ready. Cannot select model.", level="ERROR"
                )
            return

        try:
            models_data = await self.executor.get_models()
            raw_models = models_data.get("models", [])
            if not isinstance(raw_models, list):
                raw_models = []
            available_models: List[str] = []
            for model in raw_models:
                if isinstance(model, str):
                    name = model.strip()
                elif isinstance(model, dict):
                    name = str(model.get("name", "")).strip()
                else:
                    name = ""
                if name:
                    available_models.append(name)
            if not available_models:
                if chat:
                    chat.add_system_message("No models available from executor.", level="ERROR")
                return

            def update_selection(result: tuple[str, str] | None) -> None:
                if result:
                    model_id, reasoning = result
                    if target == "code":
                        self.code_model = model_id
                        self.reasoning_level_code = reasoning
                        try:
                            self.settings.last_code_model = model_id
                            self.settings.last_code_reasoning_level = reasoning
                            self.settings.save()
                        except Exception:
                            logger.exception("Failed to persist code model/reasoning settings")
                        label = "Code Model"
                    else:
                        self.current_model = model_id
                        self.reasoning_level = reasoning
                        try:
                            self.settings.last_model = model_id
                            self.settings.last_reasoning_level = reasoning
                            self.settings.save()
                        except Exception:
                            logger.exception("Failed to persist model/reasoning settings")
                        label = "Model"

                    if chat:
                        msg = f"{label}: [bold]{model_id}[/] (Reasoning: [bold]{reasoning}[/])"
                        chat.add_system_message_markup(f"Settings updated: {msg}")

                    # Update statusline (best-effort)
                    try:
                        self._update_statusline()
                    except Exception:
                        pass

            current_m = self.code_model if target == "code" else self.current_model
            current_r = self.reasoning_level_code if target == "code" else self.reasoning_level

            self.push_screen(
                ModelReasoningSelectModal(available_models, current_m, current_r),
                update_selection,
            )
        except Exception as e:
            if chat:
                chat.add_system_message(f"Failed to fetch models: {e}", level="ERROR")

    def action_select_reasoning(self, target: str = "planner") -> None:
        chat = self._maybe_chat()
        if chat:
            self._reasoning_target = target
            levels = ["disable", "low", "medium", "high"]
            current_val = self.reasoning_level_code if target == "code" else self.reasoning_level
            current = str(current_val or "low").strip() or "low"
            if current not in levels:
                current = "low"
            chat.open_reasoning_menu(levels, current)

    def action_select_mode(self) -> None:
        chat = self._maybe_chat()
        if chat:
            chat.open_mode_menu(["CODE", "ASK", "LUTZ"], self.agent_mode)

    def action_toggle_context(self) -> None:
        if isinstance(self.screen, ContextModalScreen):
            self._show_chat_token_bar()
            self.screen.dismiss(None)
            return

        chat = self._maybe_chat()
        if chat:
            chat.set_token_bar_visible(False)

        self.push_screen(ContextModalScreen(on_close=self._show_chat_token_bar))

        if self._executor_ready:
            self.run_worker(self._refresh_context_panel())

    def _show_chat_token_bar(self) -> None:
        chat = self._maybe_chat()
        if chat:
            chat.set_token_bar_visible(True)

    def _maybe_focused_widget(self) -> Any | None:
        """Best-effort focused widget lookup; returns None if there is no active screen."""
        try:
            return self.focused
        except ScreenStackError:
            return None

    def action_toggle_tasklist(self) -> None:
        """Toggle the task list modal.

        When opened, the modal's TaskListPanel receives focus.
        When closed, focus is restored to whatever had it previously (best-effort).
        """
        try:
            current_screen = self.screen
        except ScreenStackError:
            current_screen = None

        if isinstance(current_screen, TaskListModalScreen):
            self._restore_tasklist_focus()
            current_screen.dismiss(None)
            return

        self._tasklist_restore_focus_widget = self._maybe_focused_widget()

        def on_close() -> None:
            self._restore_tasklist_focus()

        self.push_screen(TaskListModalScreen(on_close=on_close))

        if self._executor_ready:
            self.run_worker(self._ensure_tasklist_data())
            self.run_worker(self._refresh_context_panel())

    def _restore_tasklist_focus(self) -> None:
        widget = self._tasklist_restore_focus_widget
        self._tasklist_restore_focus_widget = None
        if widget is None:
            return
        try:
            widget.focus()
        except Exception:
            pass

    def action_task_next(self) -> None:
        panel = self._active_tasklist_panel()
        panel.move_selection(1)

    def action_task_prev(self) -> None:
        panel = self._active_tasklist_panel()
        panel.move_selection(-1)

    def action_task_toggle(self) -> None:
        self.run_worker(self._toggle_selected_task())

    def action_task_delete(self) -> None:
        self.run_worker(self._delete_selected_task())

    def action_task_add(self) -> None:
        def on_done(result: Optional[str]) -> None:
            if result is None:
                return
            self.run_worker(self._add_task(result))

        self.push_screen(TaskTitleModalScreen("Add Task"), on_done)

    def action_task_edit(self) -> None:
        chat = self._maybe_chat()
        panel = self._active_tasklist_panel()
        selected = panel.selected_task()
        if not selected:
            if chat:
                chat.add_system_message("No task selected.")
            return

        initial = str(selected.get("title", "") or "").strip()

        def on_done(result: Optional[str]) -> None:
            if result is None:
                return
            self.run_worker(self._edit_selected_task(result))

        self.push_screen(TaskTitleModalScreen("Edit Task", initial=initial), on_done)

    def action_toggle_mode(self) -> None:
        """Cycles through agent modes: CODE -> ASK -> LUTZ -> CODE."""
        modes = ["CODE", "ASK", "LUTZ"]
        try:
            current_index = modes.index(self.agent_mode)
        except ValueError:
            current_index = 0

        next_index = (current_index + 1) % len(modes)
        self._set_mode(modes[next_index])

    def _set_theme(self, theme_name: str) -> None:
        normalized_theme = normalize_theme_name(theme_name.lower())
        resolved_theme = (
            normalized_theme if normalized_theme in self.available_themes else DEFAULT_THEME
        )
        if resolved_theme != normalized_theme:
            logger.warning(
                "Unknown theme '%s'; falling back to '%s'.",
                theme_name,
                DEFAULT_THEME,
            )
        self.theme = resolved_theme

    def watch_theme(self, old_theme: str, new_theme: str) -> None:
        """Persist any theme changes, including those from the Textual theme palette."""
        if old_theme == new_theme:
            return
        self.settings.theme = new_theme
        self.settings.save()

    async def action_handle_ctrl_c(self) -> None:
        """Handles Ctrl+C: Clear input, cancel job, or double-tap to quit."""
        chat_panel = self._maybe_chat()
        if chat_panel:
            chat_inputs = self.query("#chat-input").results(ChatInput)
            chat_input = next(chat_inputs, None)
            if chat_input and chat_input.text.strip():
                chat_panel.clear_input()
                return

        now = time.time()

        if self.job_in_progress and self.current_job_id:
            self._pending_prompt = None  # Clear any pending prompt on manual cancel
            self._pending_updated_at = 0
            self._pending_generation = 0
            self._pending_min_wait_until = 0.0
            if chat_panel:
                chat_panel.add_system_message("Cancelling job...")
            await self.executor.cancel_job(self.current_job_id)
            # Reset double-tap timer so they don't accidentally quit while cancelling
            self._last_ctrl_c_time = now
            return

        if now - self._last_ctrl_c_time < 2.0:
            await self.action_quit()
        else:
            self.query_one(ChatPanel).add_system_message("Press Ctrl+C again to quit.")
            self._last_ctrl_c_time = now

    async def _export_session(self) -> None:
        """Best-effort export of the current session zip to workspace cache.

        This method is idempotent and will only attempt export once per app lifetime.
        """
        # Avoid re-exporting multiple times
        if self._session_exported:
            return

        # Only require a session_id here; availability is checked by the caller.
        if not self.executor.session_id:
            return

        from brokk_code.session_persistence import get_session_zip_path

        try:
            session_id = self.executor.session_id
            assert session_id is not None
            zip_bytes = await self.executor.download_session_zip(session_id)
            zip_path = get_session_zip_path(self.executor.workspace_dir, session_id)
            zip_path.write_bytes(zip_bytes)
            logger.info("Session %s exported to %s", session_id, zip_path)
            self._session_exported = True
        except ExecutorError as ee:
            # Treat intentional "Executor not started" after a controlled stop as benign.
            msg = str(ee)
            if "Executor not started" in msg:
                logger.debug(
                    "Session export skipped because executor is not started "
                    "(expected during shutdown)."
                )
            else:
                logger.warning("Failed to export session zip on shutdown: %s", ee)
        except Exception as e:
            logger.warning("Failed to export session zip on shutdown: %s", e)

    async def _shutdown_once(self, *, show_message: bool = True) -> None:
        """Perform shutdown actions once completed. Concurrency-safe with retry on stop failure."""
        # Fast path
        async with self._shutdown_lock:
            if self._shutdown_completed or self._shutting_down:
                return
            self._shutting_down = True

            # Notify user once
            if show_message:
                msg = "Shutting down..."
                chat = self._maybe_chat()
                if chat:
                    chat.add_system_message(msg)
                else:
                    logger.info(msg)

            # Attempt to export the session before we stop the executor.
            # Only attempt export if executor reports ready or alive.
            try:
                can_export = False
                if self.executor.session_id:
                    # Prefer _executor_ready, but if it is false we still attempt
                    # export only if the process is alive.
                    if self._executor_ready or self.executor.check_alive():
                        can_export = True

                if can_export:
                    await self._export_session()
                else:
                    logger.debug("Skipping session export: no session or executor not available.")
            except Exception:
                logger.debug("Export step failed during shutdown", exc_info=True)

            # Mark executor not ready immediately so any concurrent refresh/export short-circuits.
            self._executor_ready = False

            # Stop executor (best-effort). Multiple calls are safe because
            # ExecutorManager.stop is idempotent.
            try:
                await self.executor.stop()
            except Exception:
                logger.debug("Executor.stop encountered an error during shutdown", exc_info=True)
                self._shutting_down = False
                return

            self._shutdown_completed = True

    async def action_quit(self) -> None:
        # Centralized shutdown; show_message True to surface to user via chat/logs.
        await self._shutdown_once(show_message=True)
        # After clean shutdown, exit the app
        try:
            self.exit()
        except Exception:
            # exit() may raise in some test harnesses; ignore to avoid double-shutdown.
            pass

    async def on_unmount(self) -> None:
        """Ensure cleanup even if app exits via other means."""
        # on_unmount is a fallback shutdown path; avoid double-running shutdown logic.
        await self._shutdown_once(show_message=False)


if __name__ == "__main__":
    app = BrokkApp()
    app.run()



import json
import stat
import sys
from pathlib import Path

import pytest

from brokk_code.zed_config import ExistingBrokkCodeEntryError, configure_zed_acp_settings


def test_configure_zed_acp_settings_creates_file(tmp_path) -> None:
    settings_path = tmp_path / ".config" / "zed" / "settings.json"

    written_path = configure_zed_acp_settings(settings_path=settings_path)

    assert written_path == settings_path
    data = json.loads(settings_path.read_text(encoding="utf-8"))
    assert data["agent_servers"]["Brokk Code"]["command"] == "brokk"
    assert data["agent_servers"]["Brokk Code"]["args"] == ["acp", "--ide", "zed"]


def test_configure_zed_acp_settings_merges_existing_values(tmp_path) -> None:
    settings_path = tmp_path / ".config" / "zed" / "settings.json"
    settings_path.parent.mkdir(parents=True, exist_ok=True)
    settings_path.write_text(
        json.dumps(
            {
                "theme": "One Dark",
                "agent_servers": {
                    "Other": {
                        "type": "custom",
                        "command": "other-agent",
                    }
                },
            }
        ),
        encoding="utf-8",
    )

    configure_zed_acp_settings(settings_path=settings_path)

    data = json.loads(settings_path.read_text(encoding="utf-8"))
    assert data["theme"] == "One Dark"
    assert "Other" in data["agent_servers"]
    assert "Brokk Code" in data["agent_servers"]


def test_configure_zed_acp_settings_rejects_existing_brokk_code(tmp_path) -> None:
    settings_path = tmp_path / ".config" / "zed" / "settings.json"
    settings_path.parent.mkdir(parents=True, exist_ok=True)
    settings_path.write_text(
        json.dumps(
            {
                "agent_servers": {
                    "Brokk Code": {
                        "command": "existing",
                    }
                }
            }
        ),
        encoding="utf-8",
    )

    with pytest.raises(ExistingBrokkCodeEntryError):
        configure_zed_acp_settings(settings_path=settings_path)


def test_configure_zed_acp_settings_force_overwrites_existing_brokk_code(tmp_path) -> None:
    settings_path = tmp_path / ".config" / "zed" / "settings.json"
    settings_path.parent.mkdir(parents=True, exist_ok=True)
    settings_path.write_text(
        json.dumps(
            {
                "agent_servers": {
                    "Brokk Code": {
                        "command": "existing",
                    }
                }
            }
        ),
        encoding="utf-8",
    )

    configure_zed_acp_settings(settings_path=settings_path, force=True)

    data = json.loads(settings_path.read_text(encoding="utf-8"))
    assert data["agent_servers"]["Brokk Code"]["command"] == "brokk"
    assert data["agent_servers"]["Brokk Code"]["args"] == ["acp", "--ide", "zed"]


def test_configure_zed_acp_settings_preserves_existing_permissions(tmp_path) -> None:
    settings_path = tmp_path / ".config" / "zed" / "settings.json"
    settings_path.parent.mkdir(parents=True, exist_ok=True)
    settings_path.write_text(json.dumps({"theme": "One Dark"}), encoding="utf-8")
    settings_path.chmod(0o640)
    expected_mode = stat.S_IMODE(settings_path.stat().st_mode)

    configure_zed_acp_settings(settings_path=settings_path)

    mode = stat.S_IMODE(settings_path.stat().st_mode)
    assert mode == expected_mode


def test_configure_zed_acp_settings_parses_jsonc(tmp_path) -> None:
    settings_path = tmp_path / ".config" / "zed" / "settings.json"
    settings_path.parent.mkdir(parents=True, exist_ok=True)
    settings_path.write_text(
        """// Zed settings
//
// Minimal example with comments and trailing commas
{
  "telemetry": {
    "diagnostics": false,
    "metrics": false,
  },
  "theme": {
    "mode": "dark",
    "light": "One Light",
    "dark": "Gruvbox Dark",
  },
}
""",
        encoding="utf-8",
    )

    configure_zed_acp_settings(settings_path=settings_path)

    updated_text = settings_path.read_text(encoding="utf-8")
    assert updated_text.startswith("// Zed settings")

    json_start = updated_text.find("{")
    assert json_start != -1
    data = json.loads(updated_text[json_start:])
    assert data["telemetry"]["diagnostics"] is False
    assert data["theme"]["dark"] == "Gruvbox Dark"
    assert data["agent_servers"]["Brokk Code"]["command"] == "brokk"


def test_configure_zed_acp_settings_default_path_linux(monkeypatch, tmp_path) -> None:
    fake_home = tmp_path / "home"
    monkeypatch.setattr(sys, "platform", "linux")
    monkeypatch.setattr(Path, "home", lambda: fake_home)
    written_path = configure_zed_acp_settings()
    assert written_path == fake_home / ".config" / "zed" / "settings.json"
    assert written_path.exists()


def test_configure_zed_acp_settings_default_path_darwin(monkeypatch, tmp_path) -> None:
    fake_home = tmp_path / "home"
    monkeypatch.setattr(sys, "platform", "darwin")
    monkeypatch.setattr(Path, "home", lambda: fake_home)
    written_path = configure_zed_acp_settings()
    assert written_path == fake_home / "Library" / "Application Support" / "Zed" / "settings.json"
    assert written_path.exists()


def test_configure_zed_acp_settings_default_path_windows_with_appdata(
    monkeypatch, tmp_path
) -> None:
    fake_appdata = tmp_path / "AppData" / "Roaming"
    monkeypatch.setattr(sys, "platform", "win32")
    monkeypatch.setenv("APPDATA", str(fake_appdata))
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    written_path = configure_zed_acp_settings()
    assert written_path == fake_appdata / "Zed" / "settings.json"
    assert written_path.exists()


def test_configure_zed_acp_settings_default_path_windows_no_appdata(monkeypatch, tmp_path) -> None:
    fake_home = tmp_path / "home"
    monkeypatch.setattr(sys, "platform", "win32")
    monkeypatch.delenv("APPDATA", raising=False)
    monkeypatch.setattr(Path, "home", lambda: fake_home)
    written_path = configure_zed_acp_settings()
    assert written_path == fake_home / "AppData" / "Roaming" / "Zed" / "settings.json"
    assert written_path.exists()


def test_configure_zed_acp_settings_default_path_windows_blank_appdata(
    monkeypatch, tmp_path
) -> None:
    fake_home = tmp_path / "home"
    monkeypatch.setattr(sys, "platform", "win32")
    monkeypatch.setenv("APPDATA", "")
    monkeypatch.setattr(Path, "home", lambda: fake_home)
    written_path = configure_zed_acp_settings()
    # Should fall back to home-based path when APPDATA is blank
    assert written_path == fake_home / "AppData" / "Roaming" / "Zed" / "settings.json"
    assert written_path.exists()


def test_configure_zed_acp_settings_default_path_windows_whitespace_appdata(
    monkeypatch, tmp_path
) -> None:
    fake_home = tmp_path / "home"
    monkeypatch.setattr(sys, "platform", "win32")
    monkeypatch.setenv("APPDATA", "   ")
    monkeypatch.setattr(Path, "home", lambda: fake_home)
    written_path = configure_zed_acp_settings()
    # Should fall back to home-based path when APPDATA is whitespace
    assert written_path == fake_home / "AppData" / "Roaming" / "Zed" / "settings.json"
    assert written_path.exists()


def test_configure_zed_acp_settings_default_path_windows_relative_appdata(
    monkeypatch, tmp_path
) -> None:
    fake_home = tmp_path / "home"
    monkeypatch.setattr(sys, "platform", "win32")
    # Relative path should be ignored
    monkeypatch.setenv("APPDATA", "Relative/Path")
    monkeypatch.setattr(Path, "home", lambda: fake_home)
    written_path = configure_zed_acp_settings()
    # Should fall back to home-based path when APPDATA is relative
    assert written_path == fake_home / "AppData" / "Roaming" / "Zed" / "settings.json"
    assert written_path.exists()


def test_configure_zed_acp_settings_default_path_windows_untrimmed_appdata(
    monkeypatch, tmp_path
) -> None:
    fake_appdata = tmp_path / "AppData" / "Roaming"
    monkeypatch.setattr(sys, "platform", "win32")
    # Path with whitespace should be trimmed and used if absolute
    monkeypatch.setenv("APPDATA", f"  {fake_appdata}  ")
    monkeypatch.setattr(Path, "home", lambda: tmp_path / "home")
    written_path = configure_zed_acp_settings()
    assert written_path == fake_appdata / "Zed" / "settings.json"
    assert written_path.exists()



from pathlib import Path
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from brokk_code.app import BrokkApp


@pytest.mark.asyncio
async def test_welcome_message_shown_on_startup(tmp_path: Path):
    """Verify welcome message appears on every startup."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path
    mock_executor.start = AsyncMock()
    mock_executor.wait_ready = AsyncMock(return_value=True)
    mock_executor.check_alive = MagicMock(return_value=True)
    mock_executor.get_health_live = AsyncMock(return_value={})
    mock_executor.create_session = AsyncMock()
    mock_executor.session_id = "test-session"
    mock_executor.stop = AsyncMock()
    mock_executor.get_context = AsyncMock(return_value={"usedTokens": 0})
    mock_executor.get_tasklist = AsyncMock(return_value={"tasks": []})

    for history in ([], ["previous prompt"]):
        with patch("brokk_code.app.load_history", return_value=history):
            app = BrokkApp(executor=mock_executor)
            async with app.run_test() as pilot:
                await pilot.pause()

                chat_log = app.query_one("#chat-log")
                content = "".join(str(line) for line in chat_log.lines)
                assert "Welcome to Brokk" in content, f"Expected welcome with history={history!r}"
                assert "Context Engineering" in content


def test_build_welcome_message_content():
    """Verify the welcome message contains expected branded content and commands."""
    from brokk_code.welcome import build_welcome_message

    mock_commands = [
        {"command": "/context", "description": "test"},
        {"command": "/task", "description": "test"},
        {"command": "/help", "description": "test"},
    ]

    msg = build_welcome_message(mock_commands)

    assert "Welcome to Brokk" in msg
    assert "context engineering" in msg.lower()
    assert "/context" in msg
    assert "/task" in msg
    assert "/help" in msg


def test_get_braille_icon_contains_braille():
    """Verify the icon helper returns characters in the Braille Unicode block."""
    from brokk_code.welcome import get_braille_icon

    icon = get_braille_icon()
    assert isinstance(icon, str)
    assert len(icon) > 0
    # Check for at least one Braille character (U+2800 to U+28FF)
    has_braille = any("\u2800" <= char <= "\u28ff" for char in icon)
    assert has_braille, "Icon should contain Unicode Braille characters"



from brokk_code.prompt_history import append_prompt, get_history_file, load_history


def test_append_and_load_history(tmp_path):
    workspace = tmp_path
    prompt = "Hello Brokk"

    append_prompt(workspace, prompt)

    history = load_history(workspace)
    assert history == [prompt]

    history_file = get_history_file(workspace)
    assert history_file.exists()
    assert ".brokk" in history_file.parts


def test_history_trimming(tmp_path):
    workspace = tmp_path
    max_n = 3

    for i in range(5):
        append_prompt(workspace, f"prompt {i}", max_history=max_n)

    history = load_history(workspace)
    assert len(history) == max_n
    assert history == ["prompt 2", "prompt 3", "prompt 4"]


def test_corrupt_history_fallback(tmp_path):
    workspace = tmp_path
    history_file = get_history_file(workspace)
    history_file.parent.mkdir(parents=True)
    history_file.write_text("not valid json")

    # Should fall back to empty list rather than crashing
    history = load_history(workspace)
    assert history == []

    # Should be able to append and recover
    append_prompt(workspace, "new prompt")
    assert load_history(workspace) == ["new prompt"]


def test_missing_directory_creation(tmp_path):
    # Ensure it creates .brokk if it doesn't exist
    workspace = tmp_path / "new_project"
    append_prompt(workspace, "test")

    assert (workspace / ".brokk").is_dir()
    assert load_history(workspace) == ["test"]
